# -*- coding: utf-8 -*-

"""
    Algorithms and calculations for the metrics used to
    identify or classify jet-stream in the literature.

    This file is in order of publish year of the metrics
    (see details_for_all_metrics.py)

    Functions and Classes ordered by the year of the paper that uses each algorithm first.
"""

# imports
from jsmetrics.utils import data_utils, windspeed_utils
import numpy as np
import scipy.ndimage
import xarray as xr

# docs
__author__ = "Thomas Keel"
__email__ = "thomas.keel.18@ucl.ac.uk"
__status__ = "Development"


def get_sum_weighted_ws(data, all_plevs_hPa):
    r"""
    Get sum of weighted windspeed.
    sum weighted windspeed is calculated as follows:

    .. math::
        \int_{p1}^{p2} (u^2+v^2)^{1/2} \,dp

    where p1, p2 is min, max pressure level

    Component of method from Koch et al (2006) https://doi.org/10.1002/joc.1255

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua' and 'va', and the coordinates: 'lon', 'lat', 'plev' and 'time'.
    all_plevs_hPa : array-like
        list of hPa unit pressure levels

    Returns
    ----------
    sum_weighted_ws : xarray.Dataset
        Data containing sum weighted windspeed values
    """
    if "plev" not in data.coords:
        raise KeyError("Data does not contain coord: 'plev'")

    if not isinstance(all_plevs_hPa, (list, np.ndarray)):
        raise TypeError("array of pressure level needs to be list or numpy.array")

    sum_weighted_ws = 0
    for plev, (i, plev_hPa) in zip(data["plev"], enumerate(all_plevs_hPa)):
        if i != 0:
            plev_hPa = plev_hPa - all_plevs_hPa[i - 1]
        sum_weighted_ws += (
            (
                data.sel(plev=plev, method="nearest")["ua"] ** 2
                + data.sel(plev=plev, method="nearest")["va"] ** 2
            )
            ** (1 / 2)
        ) * plev_hPa
    return sum_weighted_ws


def get_weighted_average_ws(sum_weighted_ws, all_plevs_hPa):
    r"""
    Calculates weighted average windspeed as follows:

    .. math::
        \alpha vel =  \frac{1}{(p2-p1)} \int_{p1}^{p2} (u^2+v^2)^{1/2} \,dp

    where p1, p2 is min, max pressure level.

    Component of method from Koch et al (2006) https://doi.org/10.1002/joc.1255

    Parameters
    ----------
    sum_weighted_ws : xarray.Dataset
        Data containing sum weighted windspeed values
    all_plevs_hPa : array-like
        list of hPa unit pressure levels

    Returns
    ----------
    weighted_average_ws : xarray.Dataset
        Data containing weighted average windspeed values
    """
    if not isinstance(all_plevs_hPa, (np.ndarray)):
        raise TypeError("array of pressure level needs to be a list or numpy.array")
    weighted_average_ws = sum_weighted_ws * (
        1 / (all_plevs_hPa.max() - all_plevs_hPa.min())
    )
    return weighted_average_ws


def get_all_hPa_list(data):
    """
    Will get a list of all the pressure levels in the data in hPa/mbar

    Parameters
    ----------
    data : xarray.Dataset
        data with plev coord

    Returns
    ----------
    plev : np.array
        arr containing pressure level list in data in hPa/mbar
    """
    if "plev" not in data.coords:
        raise KeyError("Data does not contain coord: 'plev'")

    if (
        data["plev"].units != "Pa"
        and data["plev"].units != "hPa"
        and data["plev"].units != "mbar"
        and data["plev"].units != "millibars"
    ):
        raise ValueError("Plev units need to be mbar, millibars, Pa or hPa")

    plevs = np.array([plev for plev in data["plev"]], dtype="float")
    if data["plev"].units == "Pa":
        plevs = plevs / 100
    return plevs


def get_local_jet_occurence(row, ws_threshold, u_threshold):
    r"""
    Each jet occurence is detected based on three rules applied to inputted
    wind speed (V = [u, v]):
        1. \|V\| is a local maxima in latitude and altitude plane
        2. \|V\| ≥ 30 m/s
        3. \|u\| ≥ 0 m/s.

    Component of method from Schiemann et al 2009 https://doi.org/10.1175/2008JCLI2625.1

    NOTE: will only work if 1 day is the resolution

    Parameters
    ----------
    row : xarray.Dataset
        Data of a single time unit containing windspeed (ws), plev, lat, lon
    ws_threshold : int or float
        Windspeed threshold used to extract jet events (default: 30 ms-1)
    u_threshold : int or float
        Windspeed threshold used to extract u-component wind speed (default: 0 ms^{-1})

    Returns
    ----------
    row : xarray.Dataset
        Data of a single time unit with value for jet-maxima (1 == maxima, 0 == none)

    """
    row["jet_occurence"] = (
        ("plev", "lat", "lon"),
        np.zeros((row["plev"].size, row["lat"].size, row["lon"].size)),
    )
    all_jet_occurences = []
    for lon in row["lon"]:
        current = row.sel(lon=lon, method="nearest")

        # Finding local maxima indices in the 2D array
        maxima_indices_ax0 = np.column_stack(
            data_utils.get_local_maxima(current["ws"].data, axis=0)
        )
        maxima_indices_ax1 = np.column_stack(
            data_utils.get_local_maxima(current["ws"].data, axis=1)
        )
        # find intersection of the two arrays intersection
        maxima_indices = data_utils.find_intersection_between_two_array_of_arrays(
            maxima_indices_ax0, maxima_indices_ax1
        )

        # Filter indices to remove maximas that neighbour each other (taking the first instance)
        filtered_maxima_indices = data_utils.filter_local_extremes_to_min_distance(
            maxima_indices, min_distance_threshold=2
        )
        if len(filtered_maxima_indices) > 0:
            # Creating a boolean mask for the filtered maxima
            maxima_mask = np.zeros_like(current["ws"], dtype=bool)
            maxima_mask[
                filtered_maxima_indices[:, 0], filtered_maxima_indices[:, 1]
            ] = True

            # Mask coordinates of filtered maxima
            current["jet_occurence"] = current["ws"].where(
                (maxima_mask)
                & (abs(current["ws"]) >= ws_threshold)
                & (current["ua"] > u_threshold)
            )
        else:
            # Set all values to np.nan
            current["jet_occurence"] = current["jet_occurence"].where(
                current["jet_occurence"] > 0
            )
        all_jet_occurences.append(current["jet_occurence"])
    all_jet_occurences_dataarray = xr.concat(all_jet_occurences, dim="lon")

    # Set all non-NaN values to 1
    all_jet_occurences_dataarray = (
        all_jet_occurences_dataarray / all_jet_occurences_dataarray
    )
    row["jet_occurence"] = all_jet_occurences_dataarray.transpose("plev", "lat", "lon")

    return row


def run_jet_core_and_region_algorithm_on_one_day(
    row,
    jet_core_plev_limit,
    jet_core_ws_threshold,
    jet_boundary_ws_threshold,
    ws_drop_threshold,
    jet_core_lat_distance,
    check_diagonals,
):
    r"""
    This method detects jet cores and defines a boundary region beside those cores based on two windspeed thresholds.
    Two additional checks are applied after initial detection of cores to check whether cores within the same windspeed region
    are part of the same feature (see 'jet_boundary_ws_threshold'). This is achieved by checking whether regions with multiple
    jet cores are more than a certain distance apart (see 'jet_core_lat_distance') and the windspeed between two cores does
    not drop below a threshold (see 'ws_drop_threshold'). This function runs this method on a single time unit.

    This method returns four outputs
        1. 'jet_core_mask' -- Regions within each latitude/altitude that are local maxima have windspeeds above the 'jet_core_ws_threshold'
        2. 'jet_region_mask' -- Regions above, below, left and right of the jet core with windspeed above the 'jet_boundary_ws_threshold'
        3. 'jet_region_above_ws_threshold_mask' -- All contigious regions of windspeeds emcompassing a jet core above the 'jet_boundary_ws_threshold' (i.e. not just above, below, left and right)
        4. 'ws' -- Wind speed calculated from 'ua', 'va' inputs.

    Component of method of algorithm originally introduced in Manney et al. (2011) https://doi.org/10.5194/acp-11-6115-2011

    Parameters
    ----------
    row : xarray.Dataset
        Data of single time unit containing the variables: 'ua' and 'va', and the coordinates: 'lon', 'lat', 'plev'.
    jet_core_plev_limit: tuple or array
        Sequence of two values relating to the pressure level limit of the jet cores (original paper uses (100, 400) hPa).
    jet_core_ws_threshold : int or float
        Threshold used for jet-stream core point.
    jet_boundary_ws_threshold : int or float
        Threshold for jet-stream boundary point.
    ws_drop_threshold : int or float
        Threshold for drop in windspeed along the line between cores.
    jet_core_lat_distance : int or float
        Threshold for maximum distance between cores to be counted the same.
    check_diagonals : bool
        Whether to check the diagonal edges of each maxima in the latitude-altitude plane.

    Returns
    ----------
    row : xarray.Dataset
        Data for one time unit containing four new variables (ws, jet_core_mask, jet_region_mask, jet_region_above_ws_threshold_mask)
    """

    # Step 1. Get potential cores (to later subset) using the wind speed thresholds and jet core pressure level limit.
    row["potential_jet_cores"] = row["ws"].where(
        lambda val: (val >= jet_core_ws_threshold)
        & (val.plev >= min(jet_core_plev_limit))
        & (val.plev <= max(jet_core_plev_limit)),
        0,
    )
    row["potential_jet_regions"] = row["ws"].where(
        lambda val: val >= jet_boundary_ws_threshold, 0
    )

    # Step 2. Get local maximas at each longitude
    local_maximas_dict = get_local_maximas_at_each_longitude(row, check_diagonals)

    # Step 3. Loop through the local maximas and make an mask of initial jet cores (not official as some may be in same region, see check in Step X)
    mask_shape = row["ws"].isel(lon=0).shape
    all_jet_core_mask = get_all_jet_core_mask(
        local_maximas_dict=local_maximas_dict, mask_shape=mask_shape
    )

    # Step 4. Loop through contigious regions (here known as jet region contours) and check if a jet core is within them
    all_jet_region_contour_mask = get_all_jet_region_contour_mask(
        row, local_maximas_dict=local_maximas_dict
    )

    # Step 5. Get the jet region around each local maxima only including above, below, left and right of maxima.
    all_jet_regions_mask = get_all_jet_regions_mask(
        row,
        all_jet_region_contour_mask=all_jet_region_contour_mask,
        local_maximas_dict=local_maximas_dict,
    )

    # Step 6. Remove old jet regions and define two outputs (jet_region, region_above_ws_threshold)
    row = row.drop_vars("potential_jet_regions")
    row["jet_region_mask"] = (
        ("plev", "lat", "lon"),
        np.clip(all_jet_regions_mask, 0, 1),
    )
    row["jet_region_contour_mask"] = (
        ("plev", "lat", "lon"),
        all_jet_region_contour_mask,
    )

    # Step 7. Run checks on jet cores, to check if they are part of the same jet feature
    jet_core_masks = run_checks_on_jet_cores_and_return_jet_cores(
        row,
        all_jet_core_mask,
        local_maximas_dict,
        jet_core_lat_distance,
        ws_drop_threshold,
    )

    # Step 8. Remove old and add actual jet core mask
    row = row.drop_vars("potential_jet_cores")
    row["jet_core_mask"] = (("plev", "lat", "lon"), jet_core_masks)
    return row


def get_local_maximas_at_each_longitude(row, check_diagonals):
    """
    Runs 'find_local_maxima_in_2d_dataarray' on each longitude (see docs for find_local_maxima_in_2d_dataarray for more information)

    Parameters
    ----------
    row : xr.DataArray
        Data of single time unit containing the variables: 'potential_jet_cores', and the coordinates: 'lon', 'lat', 'plev'
    check_diagonals : bool
        Whether to check the diagonal edges of each maxima in the latitude-altitude plane.

    Returns
    ----------
    local_maximas_dict : dict
        Keys are longitude coordinates, values are local maxima locations

    """
    local_maximas_dict = {}
    for _, lon in enumerate(row["lon"]):
        pot_core_one_lon = row["potential_jet_cores"].sel(lon=lon)
        if check_diagonals:
            local_maximas = find_local_maxima_in_2d_dataarray_with_diagonals(
                pot_core_one_lon
            )
        else:
            local_maximas = find_local_maxima_in_2d_dataarray(pot_core_one_lon)
        local_maximas_dict[float(lon)] = local_maximas
    return local_maximas_dict


def find_local_maxima_in_2d_dataarray(arr):
    """
    Find indices of local maximas within a 2-D array. Should return two values which relate to the position of local maximas (if any).

    Component of method of algorithm originally introduced in Manney et al. (2011) https://doi.org/10.5194/acp-11-6115-2011

    Parameters
    ----------
    arr : xr.DataArray
        A 2-D array with numeric values (i.e. dtypes float or int)

    Returns
    ----------
    local_maxima : xr.DataArray
        A 2-D array with values relating to the index of the local maximas

    Examples
    ----------
    .. code-block:: python

        import numpy as np
        import xarray as xr

        # Example array
        data = np.array([
            [1, 2, 3, 4, 5],
            [2, 3, 4, 5, 4],
            [3, 4, 5, 6, 7],
            [4, 5, 6, 7, 6],
            [5, 6, 7, 8, 9]
        ])

        # Convert NumPy array to xarray DataArray
        data_array = xr.DataArray(data)

        local_maxima_indices = find_local_maxima_in_2d_dataarray(data_array)

        for i, j in local_maxima_indices:
            print(f"Local maximum at ({i}, {j}): {data_array[i, j]}")

    """
    # Add padding to array to allow edges to be picked up
    arr = data_utils.add_pad_to_array(arr)

    # Calculate neighbors for all interior points
    neighbors = np.stack(
        [
            arr[:-2, 1:-1],  # Above
            arr[2:, 1:-1],  # Below
            arr[1:-1, :-2],  # Left
            arr[1:-1, 2:],  # Right
        ],
        axis=-1,
    )

    # Find local maximas
    interior_maxima = arr[1:-1, 1:-1] > np.max(neighbors, axis=-1)
    interior_indices = np.transpose(np.where(interior_maxima))

    return np.array(interior_indices)


def find_local_maxima_in_2d_dataarray_with_diagonals(arr):
    """
    Find indices of local maximas within a 2-D array with check for the diagonals.
    Should return two values which relate to the position of local maximas (if any).

    Component of method  from Manney et al. (2011) https://doi.org/10.5194/acp-11-6115-2011

    Parameters
    ----------
    arr : xr.DataArray
        A 2-D array with numeric values (i.e. dtypes float or int)

    Returns
    ----------
    local_maxima : xr.DataArray
        A 1-D array with values relating to the index of the local maximas

    Examples
    ----------
    .. code-block:: python

        import numpy as np
        import xarray as xr

        # Example array
        data = np.array([
            [1, 2, 3, 4, 5],
            [2, 3, 4, 5, 4],
            [3, 4, 5, 6, 7],
            [4, 5, 6, 7, 6],
            [5, 6, 7, 8, 9]
        ])

        # Convert NumPy array to xarray DataArray
        data_array = xr.DataArray(data)

        local_maxima_indices = find_local_maxima_in_2d_dataarray_with_diagonals(data_array)

        for i, j in local_maxima_indices:
            print(f"Local maximum at ({i}, {j}): {data_array[i, j]}")

    """
    local_maxima = []

    # Add padding to array to allow edges to be picked up
    arr = data_utils.add_pad_to_array(arr)

    # Calculate neighbors for all interior points
    neighbors = np.stack(
        [
            arr[:-2, 1:-1],  # Above
            arr[2:, 1:-1],  # Below
            arr[1:-1, :-2],  # Left
            arr[1:-1, 2:],  # Right
        ],
        axis=-1,
    )

    # Find local maximas
    interior_maxima = arr[1:-1, 1:-1] > np.max(neighbors, axis=-1)
    interior_indices = np.transpose(np.where(interior_maxima))
    local_maxima.extend([(i, j) for i, j in interior_indices])

    # Check edges and corners
    top_maxima = np.where(arr[1, 1:-1] > np.maximum(arr[0, 1:-1], arr[2, 1:-1]))[0]
    local_maxima.extend([(0, j) for j in top_maxima])
    bottom_maxima = np.where(arr[-2, 1:-1] > np.maximum(arr[-1, 1:-1], arr[-3, 1:-1]))[
        0
    ]

    # arr.shape[0] - 3 becuase padding added to array
    local_maxima.extend([(arr.shape[0] - 3, j) for j in bottom_maxima])

    left_maxima = np.where(arr[1:-1, 1] > np.maximum(arr[1:-1, 0], arr[1:-1, 2]))[0]
    local_maxima.extend([(i, 0) for i in left_maxima])

    right_maxima = np.where(arr[1:-1, -2] > np.maximum(arr[1:-1, -1], arr[1:-1, -3]))[0]
    local_maxima.extend([(i, arr.shape[1] - 3) for i in right_maxima])

    # Check corners
    if arr[0, 0] > np.maximum(arr[1, 0], arr[0, 1]):
        local_maxima.append((0, 0))
    if arr[0, -1] > np.maximum(arr[1, -1], arr[0, -2]):
        local_maxima.append((0, arr.shape[1] - 3))
    if arr[-1, 0] > np.maximum(arr[-2, 0], arr[-1, 1]):
        local_maxima.append((arr.shape[0] - 3, 0))
    if arr[-1, -1] > np.maximum(arr[-2, -1], arr[-1, -2]):
        local_maxima.append((arr.shape[0] - 3, arr.shape[1] - 3))

    # Remove duplicates (this strategy can produce duplicates)
    local_maxima = np.unique(np.array(local_maxima), axis=0)
    return local_maxima


def get_all_jet_core_mask(local_maximas_dict, mask_shape):
    """
    Runs 'create_mask_using_local_maximas' using each

    Component of method of algorithm originally introduced in Manney et al. (2011) https://doi.org/10.5194/acp-11-6115-2011

    Parameters
    ----------
    local_maximas_dict : dict
        Keys are longitude coordinates, values are local maxima locations

    mask_shape : tuple
        Values to create a mask of given shape from
        .
    Returns
    ----------
    all_jet_core_mask : numpy.array
        Mask of the jet cores

    """
    all_jet_core_mask = np.array([])
    for ind, local_maximas in enumerate(local_maximas_dict.values()):
        current_jet_core_mask = create_mask_using_local_maximas(
            local_maximas=local_maximas, mask_shape=mask_shape
        )
        if ind == 0:
            all_jet_core_mask = current_jet_core_mask
            continue
        all_jet_core_mask = np.dstack([all_jet_core_mask, current_jet_core_mask])
    return all_jet_core_mask


def create_mask_using_local_maximas(local_maximas, mask_shape):
    """
    Will create a mask with the same dimensions as the inputted mask_shape
    and only the local maxima as value 1. All other values will be 0.

    Component of method of algorithm originally introduced in Manney et al. (2011) https://doi.org/10.5194/acp-11-6115-2011

    Parameters
    ----------
    local_maximas : np.array
        An array containing values relating to the index of the local maximas in a mask with shape: mask_shape

    mask_shape : tuple
        Values to create a mask of given shape from.

    Returns
    ----------
    mask : np.array
        An array with shape: mask_shape with only the indexes of the local maximas as 1 All other values will 0.

    """
    empty_mask = np.zeros(mask_shape)
    for i, j in local_maximas:
        empty_mask[i, j] = 1
    return empty_mask


def get_all_jet_region_contour_mask(row, local_maximas_dict):
    """
    Runs 'get_jet_region_contour_mask' on each longitude in data with one time step

    Component of method of algorithm originally introduced in Manney et al. (2011) https://doi.org/10.5194/acp-11-6115-2011

    Parameters
    ----------
    row : xr.DataArray
        Data of single time unit containing the variables: 'potential_jet_cores', and the coordinates: 'lon', 'lat', 'plev'

    local_maximas : np.array
        An array containing values relating to the index of the local maximas in a mask with shape: mask_shape

    Returns
    ----------
    all_jet_region_contour_mask : np.array
        A 3-D array the indexes (lat, plev) of the local maximas for each longitude as 1 All other values will 0.

    """
    all_jet_region_contour_mask = np.array([])
    for ind, lon in enumerate(row.lon):
        current_jet_region_contour_mask = get_jet_region_contour_mask(
            row["potential_jet_regions"].sel(lon=lon), local_maximas_dict[float(lon)]
        )
        if ind == 0:
            all_jet_region_contour_mask = current_jet_region_contour_mask
            continue
        all_jet_region_contour_mask = np.dstack(
            [all_jet_region_contour_mask, current_jet_region_contour_mask]
        )
    return all_jet_region_contour_mask


def get_jet_region_contour_mask(potential_jet_regions, local_maximas):
    """
    Will create a mask based on regions (above ws_threshold) around the jet cores that are
    contiguous.
    The mask will contain categorical values (e.g. 1, 2) for each cluster of jet regions
    contain a jet core. All other values will be 0 (i.e not a jet core or jet region)

    Component of method of algorithm originally introduced in Manney et al. (2011) https://doi.org/10.5194/acp-11-6115-2011

    Parameters
    ----------
    potential_jet_regions : xr.DataArray
        Windspeeds of regions above the jet region threshold

    local_maximas : np.array
        An array containing values relating to the index of the local maximas (should be the same shape as input: potential_jet_regions)

    Returns
    ----------
    jet_region_contour_mask : np.array
        A 2-D array with shape: mask_shape with only the indexes of the local maximas as 1 All other values will 0.

    """
    potential_jet_regions_mask, _ = scipy.ndimage.label(potential_jet_regions)
    actual_jet_region_nums = get_jet_region_numbers(
        local_maximas, potential_jet_regions_mask
    )
    jet_region_contour_mask = subset_jet_region_mask_to_regions_with_cores(
        potential_jet_regions_mask, actual_jet_region_nums
    )

    # Refine jet regions to new values (i.e. if labels have been removed this resets them)
    jet_region_contour_mask, _ = scipy.ndimage.label(jet_region_contour_mask)
    return jet_region_contour_mask


def get_jet_region_numbers(local_maximas, potential_jet_regions_mask):
    """
    Will only return clusters ID numbers of jet regions with an actual jet core in them.

    Component of method of algorithm originally introduced in Manney et al. (2011) https://doi.org/10.5194/acp-11-6115-2011

    Parameters
    ----------
    local_maximas : np.array
        An array containing values relating to the index of the local maximas (should be the same shape as input: potential_jet_regions)

    potential_jet_regions_mask : numpy.array
        Array of jet regions clusters as returned by scipy.ndimage.label

    Returns
    ----------
    actual_jet_region_nums : list
        A list of cluster ID numbers with local maximas in them

    """
    actual_jet_region_nums = []
    for i, j in local_maximas:
        valid_region = potential_jet_regions_mask[i, j]
        if valid_region not in actual_jet_region_nums:
            actual_jet_region_nums.append(valid_region)
    return actual_jet_region_nums


def subset_jet_region_mask_to_regions_with_cores(
    potential_jet_regions_mask, actual_jet_region_nums
):
    """
    Will subset jet region mask to only those with an actual jet core in them.

    Component of method of algorithm originally introduced in Manney et al. (2011) https://doi.org/10.5194/acp-11-6115-2011

    Parameters
    ----------
    potential_jet_regions_mask : numpy.array
        Array of jet regions clusters as returned by scipy.ndimage.label

    actual_jet_region_nums : list
        A list of cluster ID numbers with local maximas in them

    Returns
    ----------
    potential_jet_regions_mask : numpy.array
        Subset version of jet regions mask containing only jet region clusters with local maxima in them

    """
    for reg_num in np.unique(potential_jet_regions_mask):
        if reg_num not in actual_jet_region_nums:
            potential_jet_regions_mask[potential_jet_regions_mask == reg_num] = 0
    return potential_jet_regions_mask


def get_all_jet_regions_mask(row, all_jet_region_contour_mask, local_maximas_dict):
    """
    Get all jet regions mask by looping over each longitude in row.
    See also docs of 'refine_jet_region_to_leftright_and_abovebelow'

    Component of method of algorithm originally introduced in Manney et al. (2011) https://doi.org/10.5194/acp-11-6115-2011

    Parameters
    ----------
    row : xr.DataArray
        Data of single time unit containing the coordinates: 'lon'

    all_jet_region_contour_mask : np.array
        A 3-D array the indexes (lat, plev) of the local maximas for each longitude as 1 All other values will 0.

    local_maximas_dict : dict
        Keys are longitude coordinates, values are local maxima locations

    Returns
    ----------
    all_jet_regions_mask : np.array
        A refined version of the inputted 'all_jet_region_contour_mask' of the local maximas for each longitude as 1 All other values will 0.

    """
    all_jet_regions_mask = np.array([])  # initialise jet region array
    for ind, lon in enumerate(row.lon):
        current_jet_region_contour_mask_one_lon = all_jet_region_contour_mask[
            ::, ::, ind
        ]
        current_jet_region_mask = np.zeros_like(
            current_jet_region_contour_mask_one_lon
        )  # create empty mask

        current_local_maximas = local_maximas_dict[float(lon)]
        for local_maxima in current_local_maximas:
            refined_result = refine_jet_region_to_leftright_and_abovebelow(
                current_jet_region_contour_mask_one_lon,
                local_maxima[0],
                local_maxima[1],
            )
            current_jet_region_mask += refined_result
        if ind == 0:
            all_jet_regions_mask = current_jet_region_mask
            continue
        all_jet_regions_mask = np.dstack(
            [all_jet_regions_mask, current_jet_region_mask]
        )
    return all_jet_regions_mask


def refine_jet_region_to_leftright_and_abovebelow(array, x, y):
    """
    This method will remove all values not left, right, above or below the input
    x, y coordinates in a 2-D array.

    Component of method of algorithm originally introduced in Manney et al. (2011) https://doi.org/10.5194/acp-11-6115-2011

    Parameters
    ----------
    array : np.array
        A 2-D array with numeric values relating to the jet region (i.e. dtypes float or int)
    x : int
        x-coordinate of a local maxima to subset jet region by
    y: int
        y-coordinate of a local maxima to subset input array by

    Returns
    ----------
    refined_array : np.array
        New 2-D array with values only around x and y coordinates

    """
    rows, cols = array.shape
    refined_array = np.zeros_like(array)

    if 0 <= x < rows and 0 <= y < cols and array[x, y] != 0:
        # Refine the current cluster
        refined_array[x, y] = array[x, y]
        left_col = y - 1
        while left_col >= 0 and array[x, left_col] != 0:
            refined_array[x, left_col] = array[x, left_col]
            left_col -= 1

        right_col = y + 1
        while right_col < cols and array[x, right_col] != 0:
            refined_array[x, right_col] = array[x, right_col]
            right_col += 1

        up_row = x - 1
        while up_row >= 0 and array[up_row, y] != 0:
            refined_array[up_row, y] = array[up_row, y]
            up_row -= 1

        down_row = x + 1
        while down_row < rows and array[down_row, y] != 0:
            refined_array[down_row, y] = array[down_row, y]
            down_row += 1

    return refined_array


def get_values_along_a_line_between_two_coordinates(data, start_point, end_point):
    """
    Get all values along a shortest path between two coordinates in a 2-D numpy array.

    Component of method of algorithm originally introduced in Manney et al. (2011) https://doi.org/10.5194/acp-11-6115-2011

    Parameters
    ----------
    arr : np.array
        A 2-D array with numeric values (i.e. dtypes float or int)

    Returns
    ----------
    local_maxima : np.array
        A 2-D array with values relating to the index of the local maximas

    Examples
    ----------
    .. code-block:: python

        import numpy as np

        # Example array
        data = np.array([
            [1, 2, 3, 4, 5],
            [2, 3, 4, 5, 4],
            [3, 4, 5, 6, 7],
            [4, 5, 6, 7, 6],
            [5, 6, 7, 8, 9]
        ])

        local_maxima_indices = find_local_maxima(data)

        for i, j in local_maxima_indices:
            print(f"Local maximum at ({i}, {j}): {data[i, j]}")

    """
    # Calculate the differences in coordinates
    dx = end_point[1] - start_point[1]
    dy = end_point[0] - start_point[0]

    # Calculate the number of steps required for the line
    num_steps = max(abs(dx), abs(dy))

    # Calculate the step sizes for each coordinate
    x_step = dx / num_steps
    y_step = dy / num_steps

    # Initialize lists to store the coordinates along the path
    path_coordinates = []
    for step in range(num_steps + 1):
        row = int(round(start_point[0] + step * y_step))
        col = int(round(start_point[1] + step * x_step))
        path_coordinates.append((row, col))

    # Extract values along the path
    values_along_path = [float(data[row, col]) for row, col in path_coordinates]

    return values_along_path


def has_ws_drop_between_cores(ws_between_cores, ws_drop_threshold):
    """
    Will check for a windspeed drop of a given threshold between cores

    Parameters
    ----------
    ws_between_cores : np.array
        A 1-D array with windspeed between two cores as determined by 'get_values_along_a_line_between_two_coordinates'
    ws_drop_threshold : int or float
        Windspeed threshold to check

    Returns
    ----------
    out : boolean
        True if windspeeds input have a drop greater than 'ws_drop_threshold'

    """
    if any(ws_between_cores[0] - np.array(ws_between_cores) > ws_drop_threshold):
        return True
    elif any(ws_between_cores[-1] - np.array(ws_between_cores) > ws_drop_threshold):
        return True
    else:
        return False


def get_index_of_cores_to_drop_based_on_multicore_regions(
    local_maxima_ind_slices, current_local_maximas, multi_core_region_ws
):
    """
    Get index of cores to remove in same region based on whether they contain the first occurence of
    the maximum windspeed or not.

    Parameters
    ----------
    local_maxima_ind_slices : np.array of np.arrays
        Collection of index slices for current local maximas
    current_local_maximas : np.array
        Collection of indexes of current local maxima
    multi_core_region_ws : xr.DataArray
        Data containing region windspeeds

    Returns
    ----------
    index_of_cores_to_drop : np.array
        Mask of jet cores in current longitude plev/lat slice with multi core regions formatted
    """
    index_of_cores_to_drop = []
    # Loop through each local_maxima_slice
    for local_maxima_ind_slice in local_maxima_ind_slices:
        if len(local_maxima_ind_slice) <= 1:
            # This would indicate a single core, so will not be removed
            continue
        # Get current max windspeed for region (as fallback). Will select first occurence of max.
        index_of_ws_maxima = get_current_region_ws_maxima_lat_and_plev_ind(
            current_local_maximas, local_maxima_ind_slice, multi_core_region_ws
        )
        previous_local_maxima = [0, 0]  # temporary
        for ind, local_maxima_ind in enumerate(local_maxima_ind_slice):
            current_local_maxima = current_local_maximas[int(local_maxima_ind)]
            if ind == 0:
                previous_local_maxima = current_local_maxima
                continue
            # Check if current or previous is the maximum (so will need to be retained)
            current_is_index_of_max = np.array_equal(
                np.array([current_local_maxima[0], current_local_maxima[1]]),
                index_of_ws_maxima,
            )
            previous_is_index_of_max = np.array_equal(
                np.array([previous_local_maxima[0], previous_local_maxima[1]]),
                index_of_ws_maxima,
            )
            # Store index of cores to drop
            if current_is_index_of_max:
                index_of_cores_to_drop.append(
                    [previous_local_maxima[0], previous_local_maxima[1]]
                )
            else:
                index_of_cores_to_drop.append(
                    [current_local_maxima[0], current_local_maxima[1]]
                )
                if not previous_is_index_of_max:
                    # occurs only if there are 3+ cores in region and maximum is not between current and previous
                    index_of_cores_to_drop.append(
                        [previous_local_maxima[0], previous_local_maxima[1]]
                    )
            previous_local_maxima = current_local_maxima
    return index_of_cores_to_drop


def get_current_region_ws_maxima_lat_and_plev_ind(
    current_local_maximas, local_maxima_ind_slice, multi_core_region_ws
):
    """
    Get the plev and lat index of current region windspeed maxima.
    Will select the first occurence of maxima, if there is more than one (rare).

    Parameters
    ----------
    current_local_maximas : np.array
        Collection of indexes of current local maxima
    local_maxima_ind_slice : np.array
        Index slice for current local maxima
    multi_core_region_ws : xr.DataArray
        Data containing region windspeeds

    Returns
    ----------
    output : np.array
        Array of the plev and latitude
    """
    all_current_local_maximas_ind = current_local_maximas[
        list(local_maxima_ind_slice.astype(int))
    ]
    all_current_local_maximas_data = multi_core_region_ws.isel(
        plev=all_current_local_maximas_ind[:, 0],
        lat=all_current_local_maximas_ind[:, 1],
    )
    if all_current_local_maximas_data.isnull().all():
        return []
    max_ws_indices = np.unravel_index(
        all_current_local_maximas_data.argmax(), all_current_local_maximas_data.shape
    )
    current_max_data = all_current_local_maximas_data[
        max_ws_indices[0], max_ws_indices[1]
    ]
    # select first occurence of maximum if multiple
    if current_max_data.size > 1:
        current_max_data = current_max_data[0]
    iplev = list(multi_core_region_ws.plev.values).index(current_max_data.plev)
    ilat = list(multi_core_region_ws.lat.values).index(current_max_data.lat)
    return np.array([iplev, ilat])


def run_checks_on_jet_cores_and_return_jet_cores(
    row,
    initial_jet_core_masks,
    local_maximas_dict,
    jet_core_lat_distance,
    ws_drop_threshold,
):
    """
    This method runs two checks on the jet cores to check whether there are regions with multiple jet cores.
    Firstly, it checks whether regions with multiple jet cores are more than a certain distance apart
    (default is 15 degrees, see 'jet_core_lat_distance'), and hence seperate cores.
    Secondly, it will check whether the windspeed between two cores drops below a threshold
    (default is 25 m/s, see 'ws_drop_threshold'), if so it will remove the latter core.

    Component of method of algorithm originally introduced in Manney et al. (2011) https://doi.org/10.5194/acp-11-6115-2011

    Parameters
    ----------
    row : xarray.Dataset
        Data of single time unit containing the variables: 'jet_region_contour_mask', and the coordinates: 'lon', 'lat', 'plev'
    initial_jet_core_masks :
        Initial mask of jet cores to check.
    ws_drop_threshold : int or float
        Threshold for drop in windspeed along the line between cores (default: 25 m/s)
    jet_core_lat_distance : int or float
        Threshold for maximum distance between cores to be counted the same (default: 15 degrees)

    Returns
    ----------
    jet_core_masks : numpy.array
        Final jet cores mask of shape of 'initial_jet_core_masks'.

    """
    jet_core_masks = np.copy(initial_jet_core_masks)
    for lon_ind, lon in enumerate(row.lon):
        jet_region_contour_one_lon = row["jet_region_contour_mask"].sel(lon=lon)
        ws_one_lon = row["ws"].sel(lon=lon)
        current_local_maximas = local_maximas_dict[float(lon)]
        # Get first occurence of the maximum windspeed
        index_of_first_maximum = np.dstack(np.where(ws_one_lon == ws_one_lon.max()))[0]

        core_and_location = []
        for core_ind, local_maxima in enumerate(current_local_maximas):
            region_within = jet_region_contour_one_lon[
                local_maxima[0], local_maxima[1]
            ]  # this is the region contour that the jet core is found within
            core_and_location.append([core_ind, float(region_within)])
        core_and_location = np.array(core_and_location)

        if len(core_and_location) == 0:
            # No cores found, so selecting the cell with maximum wind speed for this lon slice
            core_and_location = index_of_first_maximum

        region_ind, num_cores_in_region = np.unique(
            core_and_location[::, 1], return_counts=True
        )
        multi_core_regions = region_ind[num_cores_in_region > 1]

        if len(multi_core_regions) == 0:
            # no multi-core regions found
            continue

        for multi_core_region in multi_core_regions:
            # get all the local maxima within regions of multi cores
            local_maxima_inds = core_and_location[::, 0][
                core_and_location[::, 1] == multi_core_region
            ]

            # Check 1. Test if cores are more than 15 degrees away
            previous_lat = None
            for local_maxima_ind in local_maxima_inds:
                local_maxima = current_local_maximas[int(local_maxima_ind)]
                if not previous_lat:
                    # set previous latitude to check for latitude distance in next
                    previous_lat = local_maxima[1]
                else:
                    current_lat = local_maxima[1]
                    if (
                        abs(row["lat"][previous_lat] - row["lat"][current_lat])
                        > jet_core_lat_distance
                    ):
                        previous_lat = current_lat
                        continue

            # Check 2. Test if cores have ws drop between them
            multi_core_region_ws = ws_one_lon.where(
                jet_region_contour_one_lon == multi_core_region
            )
            for multi_core_region in multi_core_regions:
                # get all the local maxima within regions of multi cores
                local_maxima_inds = core_and_location[::, 0][
                    core_and_location[::, 1] == multi_core_region
                ]
                # Step 1 of 2: loop over local maxima in multicore region and find if they need to split further
                slicepoints_in_region = []
                previous_local_maxima = [0, 0]  # temporary
                for ind, local_maxima_ind in enumerate(local_maxima_inds):
                    current_local_maxima = current_local_maximas[int(local_maxima_ind)]
                    if ind == 0:
                        previous_local_maxima = current_local_maxima
                        continue
                    windspeeds_between_cores = (
                        get_values_along_a_line_between_two_coordinates(
                            multi_core_region_ws,
                            start_point=previous_local_maxima,
                            end_point=current_local_maxima,
                        )
                    )
                    # Get conditions for dropping cores
                    ws_drops_below_threshold = has_ws_drop_between_cores(
                        windspeeds_between_cores, ws_drop_threshold=ws_drop_threshold
                    )
                    if ws_drops_below_threshold:
                        slicepoints_in_region.append(ind)
                    previous_local_maxima = current_local_maxima
                # Step 2 of 2: Remove cores in same region and retain core with maximum windspeed
                local_maxima_ind_slices = data_utils.slice_array_by_index_breaks(
                    local_maxima_inds, slicepoints_in_region
                )
                index_of_cores_to_drop = (
                    get_index_of_cores_to_drop_based_on_multicore_regions(
                        local_maxima_ind_slices,
                        current_local_maximas,
                        multi_core_region_ws,
                    )
                )
                for index_of_core_to_drop in index_of_cores_to_drop:
                    jet_core_masks[
                        index_of_core_to_drop[0], index_of_core_to_drop[1], lon_ind
                    ] = 0
    return jet_core_masks


def get_empty_local_wind_maxima_data(
    data, expected_dims=("time", "plev", "lat", "lon")
):
    """
    Will add a new data var of zeros for local wind maxima

    Component of method from Pena-Ortiz (2013) https://doi.org/10.1002/jgrd.50305

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua' and 'va', and the expected dims e.g.: 'lon', 'lat', 'plev' and 'time'.

    expected_dims : tuple
        Expected dimensions of input data (default: "time", "plev", "lat", "lon")

    Returns
    ----------
    data : xarray.Dataset
        Data containing zeros array of (time, plev, lat, lon) dimensions
    """
    expected_dims_set = set(expected_dims)
    data_dim_set = set(data.dims)
    actual_dims = tuple(expected_dims_set.intersection(data_dim_set))
    dim_sizes = tuple(len(data[coord]) for coord in actual_dims)
    data["local_wind_maxima"] = (
        actual_dims,
        np.zeros(dim_sizes),
    )
    return data


def get_potential_local_wind_maximas_by_ws_threshold(ws_slice, ws_threshold=30):
    """
    Will return a 2-d array of potential local windspeed maximas

    Component of method from Pena-Ortiz (2013) https://doi.org/10.1002/jgrd.50305

    Parameters
    ----------
    ws_slice : xarray.Dataset
        Data slice of windspeed that has only 'lat' and 'lon' dims

    ws_threshold : int or float
        windspeed threshold to apply (default=30 ms-1)
    Returns
    ----------
    ws_slice : xarray.Dataset
        Data slice of windspeed (lat, lon only) with ws_threshold applied
    """
    return ws_slice.where(lambda x: x > ws_threshold).fillna(0.0)


def get_local_wind_maxima_by_timeunit(row, ws_threshold):
    """
    Get local wind maxima by timeunit (i.e. day)

    Component of method from Pena-Ortiz (2013) https://doi.org/10.1002/jgrd.50305 who originally use 30 m/s as their ws threshold

    Parameters
    ----------
    row : xarray.Dataset
        Data of single time unit containing the variables: 'ua' and 'va', and the coordinates: 'lon', 'lat', 'plev'

    ws_threshold : int or float
        windspeed threshold to apply

    Returns
    ----------
    row : xarray.Dataset
        Data of a single time unit containing 0 or 1 value (local wind maxima) for that time unit
    """
    if "local_wind_maxima" not in row.data_vars:
        raise ValueError("local_wind_maxima needs to be defined.")

    row = row.transpose("plev", "lat", ...)
    for lon in row["lon"]:
        current = row.sel(lon=lon, method="nearest")
        pot_local_maximas = get_potential_local_wind_maximas_by_ws_threshold(
            current["ws"], ws_threshold
        ).data
        ind_local_wind_maximas = data_utils.get_local_maxima(pot_local_maximas, axis=1)
        # Turn into 2-d numpy array
        ind_local_wind_maximas = np.array(
            [
                [arr1, arr2]
                for arr1, arr2 in zip(
                    ind_local_wind_maximas[0], ind_local_wind_maximas[1]
                )
            ]
        )
        for plev_ind, lat_ind in ind_local_wind_maximas:
            row["local_wind_maxima"].loc[
                dict(
                    plev=current["plev"].data[plev_ind],
                    lon=lon,
                    lat=current["lat"].data[lat_ind],
                )
            ] = 1.0
    return row


def get_number_of_timeunits_per_monthyear_with_local_wind_maxima(data):
    """
    Will resample by each month and return number of timeunits (i.e. day) with local wind maxima.

    Component of method from Pena-Ortiz (2013) https://doi.org/10.1002/jgrd.50305

    Parameters
    ----------
    data : xarray.Dataset
        Input data with local wind maxima values by time unit (i.e. day)

    Returns
    ----------
    data : xarray.Dataset
        Data containing zeros array of (time, plev, lat, lon) dimensions

    """
    data["local_wind_maxima_by_monthyear"] = (
        data["local_wind_maxima"]
        .resample(time="MS")
        .sum()
        .rename({"time": "monthyear"})
    )
    return data


def subdivide_local_wind_maxima_into_stj_pfj(
    data, local_wind_column_name="local_wind_maxima_by_monthyear"
):
    """
    Subdivide the local_wind_maxima values into the Subtropical Jet (STJ) and Polar Front Jet (PFJ) based on Table 1 pg. 2709
    from Pena-Ortiz (2013) https://doi.org/10.1002/jgrd.50305.
    After the method in that paper, categorisation for the Northern Hemisphere STJ is only possible in DJF,
    and the latitude bands used are not based on December, March, June or September in Table 1 of that study.

    Component of method from Pena-Ortiz (2013) https://doi.org/10.1002/jgrd.50305

    Parameters
    ----------
    data : xarray.Dataset
        Input data with local wind maxima values

    Returns
    ----------
    data : xarray.Dataset
        data with polar_front_jet and subtropical_jet subdivisions
    """
    DJF_STJ = data.sel(
        monthyear=data.monthyear.dt.month.isin([12, 1, 2]), lat=slice(15, 40)
    )[local_wind_column_name]
    MAM_SON_PFJ = data.sel(
        monthyear=data.monthyear.dt.month.isin([3, 4, 5, 9, 10, 11]),
        lat=slice(10, 70),
    )[local_wind_column_name]
    JJA_PFJ = data.sel(
        monthyear=data.monthyear.dt.month.isin([6, 7, 8]), lat=slice(30, 60)
    )[local_wind_column_name]
    SH_STJ = data.sel(lat=slice(-40, -15))[local_wind_column_name]
    SH_PFJ = data.sel(lat=slice(-70, -41))[local_wind_column_name]
    data["polar_front_jet"] = xr.merge([MAM_SON_PFJ, JJA_PFJ, SH_PFJ])[
        local_wind_column_name
    ]
    data["subtropical_jet"] = xr.merge([DJF_STJ, SH_STJ])[local_wind_column_name]
    return data


def run_jet_occurence_and_centre_alg_on_one_day(row, occurence_ws_threshold):
    """
    Runs JetStreamCoreIdentificationAlgorithm method on a single day.

    Component of method from Kuang et al (2014) https://doi.org/10.1007/s00704-013-0994-x

    Parameters
    ----------
    row : xarray.Dataset
        Data of single time unit containing the variables: 'ua' and 'va', and the coordinates: 'lon', 'lat', 'plev'
    occurence_ws_threshold : int or float
        Threshold used to identify a jet-stream occurence point

    Returns
    ----------
    occ_alg.output_data : xarray.Dataset
        Data with jet occurence and centre points (1 for occurence, 2 for centre)
    """
    occ_alg = JetStreamOccurenceAndCentreAlgorithm(row, occurence_ws_threshold)
    occ_alg.run()
    return occ_alg.output_data


class JetStreamOccurenceAndCentreAlgorithm:
    """
    Jet-stream occurence and centre algorithm.

    Component of method from Kuang et al (2014) https://doi.org/10.1007/s00704-013-0994-x
    """

    def __init__(self, data, occurence_ws_threshold=30):
        """
        Parameters
        ----------
        data : xarray.Dataset
           Data of single time unit containing the variables: 'ua' and 'va', and the only the coordinates: 'lon', 'lat'
        occurence_ws_threshold : int or float
            Threshold used to identify a jet-stream occurence point
        """
        try:
            assert occurence_ws_threshold > 0
        except Exception as e:
            raise ValueError(
                "Occurence wind-speed threshold needs to be more than 0"
            ) from e

        # Load in data as a pressure level 2d wind-speed slice
        self.plev_ws_slice = windspeed_utils.PressureLevelWindSpeedSlice(data).values
        self.plev_ws_slice["jet_ocurrence1_jet_centre2"] = self.plev_ws_slice[
            "ws"
        ].copy()
        self.plev_ws_slice["jet_ocurrence1_jet_centre2"] = self.plev_ws_slice[
            "jet_ocurrence1_jet_centre2"
        ].where(lambda x: x >= occurence_ws_threshold)
        self._jet_occurence = self.plev_ws_slice
        self._lat_resolution = abs(
            float(self.plev_ws_slice["lat"][1] - self.plev_ws_slice["lat"][0])
        )
        self._lon_resolution = abs(
            float(self.plev_ws_slice["lon"][1] - self.plev_ws_slice["lon"][0])
        )

        # make_duplicate data for output
        self.output_data = self._jet_occurence.copy(deep=True)

        # Initialise lists needed for search algorithm
        self._all_coords = []
        self._lats_with_3 = []
        self._lats_for_search = []
        self._jet_centres = []

        # Needed to keep track of algorithm
        self.algorithm_has_run = False

    @classmethod
    def run_algorithm(cls, data):
        """
        Class method for running the algorithm.
        """
        alg_output = cls(data)
        alg_output.run()
        return alg_output

    def run(self):
        """
        Run the algorithm.
        """
        self._get_all_coords_of_jet_occurence()
        self._all_coords_arr = np.array(self._all_coords)
        self._calc_jet_centre_points()
        self._get_jet_centre_data()
        self._label_jet_occurence()

        self.algorithm_has_run = True

    def _get_jet_centre_data(self):
        """
        Calculates jet-stream centres based on if one jet-stream occurence grid
        is surrounded by 8 cells of jet-stream occurence (default is 30 m/s)
        """
        # to fix: there's got to be a quicker way
        for centre in self._jet_centres:
            self.output_data["jet_ocurrence1_jet_centre2"].loc[
                dict(lat=centre[0], lon=centre[1])
            ] = 2

    def _get_all_coords_of_jet_occurence(self):
        """
        Get all coords with the jet occurence value.
        """
        for val in self._jet_occurence["jet_ocurrence1_jet_centre2"].notnull():
            if val.any():
                for sub_val in val:
                    if sub_val:
                        self._all_coords.append(
                            [float(sub_val["lat"]), float(sub_val["lon"])]
                        )

    def _get_all_latitudes_available_in_3by3_grid(self):
        """
        Get all latitudes available in 3 by 3 grid.
        """
        for lat in self._lats_with_3:
            if (
                lat - self._lat_resolution in self._lats_with_3
                and lat + self._lat_resolution in self._lats_with_3
            ):
                self._lats_for_search.append(lat)

    def _calc_jet_centre_points(self):
        """
        Will return a list of the coordinates for all jet-centre points
        """
        for coord in self._all_coords_arr:
            coord_ws = float(
                self.output_data.sel(lat=coord[0], lon=coord[1], method="nearest")["ws"]
            )
            lat_grid_vals = np.arange(
                coord[0] - self._lat_resolution,
                coord[0] + self._lat_resolution + 0.01,
                self._lat_resolution,
            )
            lat_grid_vals = lat_grid_vals[
                (lat_grid_vals >= float(self.output_data["lat"].min()))
                & (lat_grid_vals <= float(self.output_data["lat"].max()))
            ]
            lon_grid_vals = np.arange(
                coord[1] - self._lon_resolution,
                coord[1] + self._lon_resolution + 0.01,
                self._lon_resolution,
            )
            lon_grid_vals = lon_grid_vals % 360  # loop lon around
            matrix_vals_to_check = np.array(
                np.meshgrid(lat_grid_vals, lon_grid_vals)
            ).T.reshape(-1, 2)
            add_coord = True
            for val_to_check in matrix_vals_to_check:
                if val_to_check[-1] not in self.output_data["lon"]:
                    continue
                if (
                    float(
                        self.output_data.sel(
                            lat=val_to_check[0], lon=val_to_check[1], method="nearest"
                        )["ws"]
                    )
                    > coord_ws
                ):
                    add_coord = False
                    break
            if add_coord:
                self._jet_centres.append(coord)

    def _label_jet_occurence(self):
        """
        Will label all non 2 values of windspeed for the occurence
        Used in Kuang et al. 2014
        """
        self.output_data["jet_ocurrence1_jet_centre2"] = self.output_data[
            "jet_ocurrence1_jet_centre2"
        ].where(lambda x: np.isfinite(x), 0)
        self.output_data["jet_ocurrence1_jet_centre2"] = self.output_data[
            "jet_ocurrence1_jet_centre2"
        ].where(lambda x: ((x == 0) | (x == 2)), 1)


def run_jet_core_algorithm_on_one_day(row, ws_core_threshold, ws_boundary_threshold):
    """
    Runs JetStreamCoreIdentificationAlgorithm method on a single time unit.

    Component of method of jet_core_identification_algorithm in jsmetrics and is inspired by
    the method from Manney et al. (2011) (https://doi.org/10.5194/acp-11-6115-2011),
    which is described in Section 3.1 of that study.

    Parameters
    ----------
    row : xarray.Dataset
        Data of single time unit containing the variables: 'ua' and 'va', and the coordinates: 'lon', 'lat', 'plev'

    ws_core_threshold : int or float
        Threshold used for jet-stream core point
    ws_boundary_threshold : int or float
        Threshold for jet-stream boundary point

    Returns
    ----------
    row : xarray.Dataset
        Data for one time unit containing jet-cores (ID number relates to each unique core)
    """
    row["jet_core_id"] = (
        ("plev", "lat", "lon"),
        np.zeros((row["plev"].size, row["lat"].size, row["lon"].size)),
    )
    for lon in row["lon"]:
        current = row.sel(lon=lon, method="nearest")
        core_alg = JetStreamCoreIdentificationAlgorithm(
            current,
            ws_core_threshold=ws_core_threshold,
            ws_boundary_threshold=ws_boundary_threshold,
        )
        core_alg.run()
        row["jet_core_id"].loc[dict(lon=lon)] = core_alg.output_data["core_id"]
    return row


class JetStreamCoreIdentificationAlgorithm:
    """
    Jet-stream core identification algorithm.

    Component of method of jet_core_identification_algorithm in jsmetrics and is inspired by
    the method from Manney et al. (2011) (https://doi.org/10.5194/acp-11-6115-2011),
    which is described in Section 3.1 of that study.

    Methods
    -------
    run:
        run algorithm
    run_algorithm:
        class method for running algorithm
    """

    def __init__(self, data, ws_core_threshold=40, ws_boundary_threshold=30):
        """
        Input will need to be longitudinal slice of windspeed values.

        Component of method of jet_core_identification_algorithm in jsmetrics and is inspired by
        the method from Manney et al. (2011) (https://doi.org/10.5194/acp-11-6115-2011),
        which is described in Section 3.1 of that study.


        Parameters
        ----------
        data : xarray.Dataset
            Data of single time unit containing the variables: 'ua' and 'va', and the coordinates: 'lon', 'lat', 'plev'
        ws_core_threshold : int or float
            Threshold used for jet-stream core point (default=40)
        ws_boundary_threshold : int or float
            Threshold for jet-stream boundary point (default=30)
        """
        try:
            assert (
                ws_core_threshold > ws_boundary_threshold
                and ws_core_threshold > 0
                and ws_boundary_threshold > 0
            )
        except Exception as e:
            raise ValueError(
                "Windspeed core threshold needs to be more than boundary\
                    threshold and both need to be more than 0"
            ) from e
        # Transpose data
        data = data.transpose(*(..., "lat", "plev"))

        # Step 1. make windspeed slice
        self._lat_ws_slice = windspeed_utils.LatitudeWindSpeedSlice(data)

        # Step 2. Get core and potential boundary points
        self._labelled_data = self._lat_ws_slice.label_slice(
            self._lat_ws_slice["ws"] < ws_core_threshold, "Core"
        )
        self._labelled_data = self._labelled_data.where(
            (self._lat_ws_slice["ws"] < ws_boundary_threshold)
            | (self._lat_ws_slice["ws"] > ws_core_threshold),
            other="Potential Boundary",
        )

        # Step 3. Get indexes of jet-stream cores and potential boundaries
        (
            self._initial_core_ids,
            self._pot_boundary_ids,
        ) = self._get_indexes_of_core_and_boundaries()

        # Step 4. Set variables needed to keep track of
        self._current_core_lat = -1
        self._currently_a_core = None
        self.algorithm_has_run = False

    def __repr__(self):
        """
        Representation of the class. Have it return the labelled data
        """
        if not self.algorithm_has_run:
            print(
                "A total of %d initial Jet-stream cores have been found\
                 in the wind-speed slice"
                % (self._labelled_data["ws"].where(lambda x: x == "Core").count())
            )
            print(
                "A total of %d potential Jet-stream boundaries have been found\
                 in the wind-speed slice"
                % (
                    self._labelled_data["ws"]
                    .where(lambda x: x == "Potential Boundary")
                    .count()
                )
            )
            return repr(self._labelled_data)
        else:
            print("A total of %d cores have been discovered" % (self.num_of_cores))
            return repr(self.output_data)

    @classmethod
    def run_algorithm(cls, data, ws_core_threshold=40, ws_boundary_threshold=30):
        """
        Class method for running algorithm
        """
        js_algorithm = cls(
            data,
            ws_core_threshold=ws_core_threshold,
            ws_boundary_threshold=ws_boundary_threshold,
        )

        js_algorithm.run()
        return js_algorithm

    def run(self):
        """
        Runs algorithm.
        """
        self.final_jet_cores = self._get_jet_core_boundary()
        self.output_data = self._add_cores_to_data()
        self.algorithm_has_run = True

    def _get_indexes_of_core_and_boundaries(self):
        """
        Will return the indexes in the ws data that ARE jet-stream cores
        and COULD BE jet-stream core boundaries
        """
        pot_boundary_ids = np.where(self._labelled_data["ws"] == "Potential Boundary")
        initial_core_ids = np.where(self._labelled_data["ws"] == "Core")
        pot_boundary_ids = np.stack(pot_boundary_ids, axis=-1)
        initial_core_ids = np.stack(initial_core_ids, axis=-1)
        return initial_core_ids, pot_boundary_ids

    @staticmethod
    def _get_indexes_to_check(pot_boundary):
        """
        Will return an array of indexes to check for potential boundaries
        or jetstream cores.
        """
        vals_to_check = []
        if pot_boundary[0] != 0:
            vals_to_check.append([pot_boundary[0] - 1, pot_boundary[1]])
        vals_to_check.append([pot_boundary[0] + 1, pot_boundary[1]])
        if pot_boundary[1] != 0:
            vals_to_check.append([pot_boundary[0], pot_boundary[1] - 1])
        vals_to_check.append([pot_boundary[0], pot_boundary[1] + 1])
        return vals_to_check

    def _get_pot_jetcore_area(self, vals, area, core_found=False):
        """
        Recursive function that will return the IDs of a jet core boundary
        i.e. above 30 m/s surrounding a core of 40 m/s
        Will check one an area of potential boundaries contains a core
        and thus can be called boundaries.
        """
        vals_copy = vals.copy()
        for val in vals:
            if val in area:
                continue
            if val in self._initial_core_ids.tolist():
                core_found = True
                # look for a new core if it is more than 15 degrees away
                if (
                    val[1] - self._current_core_lat > 15
                    and val[1] > self._current_core_lat
                ):
                    #  print('THIS IS A NEW CORE')
                    self._current_core_lat = val[1]
                area.append(val)
                new_vals = self._get_indexes_to_check(val)
                vals_copy.extend(new_vals)
                vals_copy = data_utils.remove_duplicates(vals_copy)
                return self._get_pot_jetcore_area(
                    vals_copy, area=area, core_found=core_found
                )

            elif val in self._pot_boundary_ids.tolist():
                area.append(val)
                new_vals = self._get_indexes_to_check(val)
                vals_copy.extend(new_vals)
                vals_copy = data_utils.remove_duplicates(vals_copy)
                return self._get_pot_jetcore_area(
                    vals_copy, area=area, core_found=core_found
                )
            else:
                vals_copy.remove(val)
                continue

        # reset current core variables
        self._current_core_lat = -1
        self._currently_a_core = None
        return area, core_found

    def _get_jet_core_boundary(self):
        """
        Recursive function that will return the IDs of all jet core boundaries
        i.e. above 30 m/s surrounding a core of 40 m/s
        Will check if an area of potential boundaries contains a core
        and thus can be called boundaries.
        """
        already_covered = []
        js_core_indexes = []
        id_number = 0
        for pot_boundary in self._pot_boundary_ids:
            if pot_boundary.tolist() in already_covered:
                continue
            vals_to_check = self._get_indexes_to_check(pot_boundary)
            area, core_found = self._get_pot_jetcore_area(vals_to_check, area=[])
            already_covered.extend(area)
            already_covered = data_utils.remove_duplicates(already_covered)
            # add area to js_core_indexes if part of core
            if core_found:
                id_number += 1
                js_core_indexes.extend([{"id": id_number, "index_of_area": area}])
        self.num_of_cores = id_number
        return js_core_indexes

    def _add_cores_to_data(self):
        """
        Add cores to data.
        """
        self._lat_ws_slice.values["core_id"] = (
            ("plev", "lat"),
            np.zeros(
                (
                    self._lat_ws_slice.values["plev"].size,
                    self._lat_ws_slice.values["lat"].size,
                )
            ),
        )
        for jet_core in self.final_jet_cores:
            for lat, plev in jet_core["index_of_area"]:
                self._lat_ws_slice.values["core_id"].loc[
                    dict(
                        lat=self._lat_ws_slice.values["lat"].data[lat],
                        plev=self._lat_ws_slice.values["plev"].data[plev],
                    )
                ] = jet_core["id"]
        return self._lat_ws_slice.values



# -*- coding: utf-8 -*-

"""
    Algorithms and calculations for the metrics used to
    identify or classify jet-stream in the literature.

    This file is in order of publish year of the metrics
    (see details_for_all_metrics.py)

    Functions and Classes ordered by the year of the paper that uses each metric component first
"""

# imports
import warnings
import numpy as np
import xarray as xr
import scipy.fftpack
import scipy.interpolate
import scipy.optimize
from jsmetrics.utils import data_utils

# docs
__author__ = "Thomas Keel"
__email__ = "thomas.keel.18@ucl.ac.uk"
__status__ = "Development"


def calc_atmospheric_mass_at_kPa(pressure, gravity=9.81, atmospheric_area=5.1e8):
    """
    Will calculate the atmospheric mass at a given pressure level.

    Radius of earth (R) = 6.372E3 km;
    Surface area of earth = 4 Pi R^2 = 5.1E8 km^2

    Component of method from Archer & Caldiera (2008) https://doi.org/10.1029/2008GL033614
    & Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1

    Parameters
    ---------------
    pressure : float
        in kPa
    gravity : float
        m/s^2
    """
    return (pressure / gravity) * atmospheric_area


def get_atm_mass_at_one_hPa(hPa):
    """
    Get atmospheric mass at one pressure level.

    Component of method from Archer & Caldiera (2008) https://doi.org/10.1029/2008GL033614
    & Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1

    Parameters
    ----------
    hPa : int or float
        One pressure level in hPa

    Returns
    ----------
    atm_mass : int or float
        Atmospheric mass at given hPa pressure level
    """
    kPa = hPa / 10
    atm_mass = calc_atmospheric_mass_at_kPa(kPa)
    return atm_mass


def get_weighted_average_at_one_pressure_level(data, pressure_level, atm_mass, ws_col):
    """
    Get weighted average wind speed at one pressure level.

    Component of method from Archer & Caldiera (2008) https://doi.org/10.1029/2008GL033614
    & Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1

    Parameters
    ----------
    data : xarray.Dataset
        Data containing windspeed ('ws') or u-, v-component wind
    pressure_level : int or float
        Pressure level to subset
    atm_mass : int or float
        Atmospheric mass at given hPa pressure level
    ws_col : string
        Name of column to calculate weighted average from (e.g. 'ws', 'ua', 'va')

    Returns
    ----------
    output : xarray.Dataset
        Data with weighted average at a single pressure level
    """
    return atm_mass * (data[ws_col].sel(plev=pressure_level))


def get_mass_weighted_average_wind(data, ws_col, plev_flux=False):
    """
    Get mass-weighted average wind-speed from 'u', 'v' component wind or 'wind vector'.

    Component of method from Archer & Caldiera (2008) https://doi.org/10.1029/2008GL033614
    & Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1

    Parameters
    ----------
    data : xarray.Dataset
        Data containing u- and v-component wind
    ws_col : string
        Name of column to calculate weighted average from (e.g. 'ws', 'ua', 'va')

    Returns
    ----------
    output : xarray.Dataset
        Data containing mass weighted average ws

    Usage
    ----------
        mon_mean = data.groupby("time.month").mean()
        mass_weighted_average = jetstream_metrics_utils.get_mass_weighted_average_ws(mon_mean)
    """
    data_utils.check_at_least_n_plevs_in_data(data, n_plevs=2)
    sum_weighted_ws = None
    (
        hPa_multiplying_factor,
        _,
    ) = get_hPa_or_mbar_and_Pa_or_bar_multiplying_factors_for_plev(data)
    for plev_step in data["plev"].data:
        plev_hPa = plev_step * hPa_multiplying_factor
        atm_mass = get_atm_mass_at_one_hPa(plev_hPa)
        weighted_average = get_weighted_average_at_one_pressure_level(
            data, plev_step, atm_mass, ws_col
        )
        if sum_weighted_ws is None:
            if plev_flux:
                sum_weighted_ws = weighted_average * plev_hPa
            else:
                sum_weighted_ws = weighted_average
        else:
            if plev_flux:
                sum_weighted_ws += weighted_average * plev_hPa
            else:
                sum_weighted_ws += weighted_average
    return sum_weighted_ws


def get_sum_atm_mass(data):
    """
    Sum atmospheric mass for given pressure levels.

    Component of method from Archer & Caldiera (2008) https://doi.org/10.1029/2008GL033614
    & Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1

    Parameters
    ----------
    data : xarray.Dataset
        Data containing plev

    Returns
    ----------
    sum_atm_mass : int or float
        Sum of atmospheric mass
    """
    sum_atm_mass = 0
    data_utils.check_at_least_n_plevs_in_data(data, n_plevs=2)
    (
        hPa_multiplying_factor,
        _,
    ) = get_hPa_or_mbar_and_Pa_or_bar_multiplying_factors_for_plev(data)
    for plev_level in data["plev"].data:
        plev_hPa = plev_level * hPa_multiplying_factor
        atm_mass = get_atm_mass_at_one_hPa(plev_hPa)
        sum_atm_mass += atm_mass
    return sum_atm_mass


def get_hPa_or_mbar_and_Pa_or_bar_multiplying_factors_for_plev(data):
    """
    Will get the multiplying factor for plev units.
    E.g. if data contains plev with the unit: Pa the two multiplying factors will be 0.01 and 1
    E.g. if data contains plev with the unit: hPa the two multiplying factors will be 1 and 100

    Component of method from Archer & Caldiera (2008) https://doi.org/10.1029/2008GL033614
    & Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1

    Parameters
    ----------
    data : xr.DataArray or xr.Dataset
        Data containing plev coord and units for plev

    Returns
    ----------
    hPa_multiplying_factor : int
        Factor to multiply plev units to convert them to hPa
    Pa_multiplying_factor : int
        Factor to multiply plev units to convert them to Pa

    """
    hPa_multiplying_factor = 1
    Pa_multiplying_factor = 1
    plev_units = data_utils.check_plev_units(
        data, ["Pa", "hPa", "mbar", "millibars", "bars"]
    )

    if plev_units in ["hPa", "mbar", "millibars"]:
        Pa_multiplying_factor *= 100
    else:
        hPa_multiplying_factor /= 100

    return hPa_multiplying_factor, Pa_multiplying_factor


def calc_mass_weighted_average(data, ws_col):
    """
    Calculate mass weighted average wind-speed.
    Component of method from Archer & Caldiera (2008) https://doi.org/10.1029/2008GL033614
    & Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1

    Parameters
    ----------
    data : xarray.Dataset
        Data containing plev
    ws_col : string
        Name of column to calculate weighted average from (e.g. 'ws', 'ua', 'va')

    Returns
    ----------
    weighted_average : xr.DataArray
        Data with weighted average windspeed based on sum atmospheric mass
    """
    sum_atm_mass = get_sum_atm_mass(data)
    sum_weighted_ws = get_mass_weighted_average_wind(data, ws_col)
    weighted_average = sum_weighted_ws / sum_atm_mass
    return weighted_average


def calc_mass_flux_weighted_pressure(data, ws_col):
    """
    Calculate mass-flux weighted pressure.

    Component of method from Archer & Caldiera (2008) https://doi.org/10.1029/2008GL033614

    Parameters
    ----------
    data : xarray.Dataset
        Data containing windspeed and plev
    ws_col : string
        Name of column to calculate weighted average from (e.g. 'ws', 'ua', 'va')

    Returns
    ----------
    mass_flux_weighted_pressure : xr.DataArray
        Data with mass flux weighted pressure
    """
    sum_weighted_ws = get_mass_weighted_average_wind(data, ws_col=ws_col)
    sum_weighted_ws_plev_flux = get_mass_weighted_average_wind(
        data, ws_col=ws_col, plev_flux=True
    )
    mass_flux_weighted_pressure = sum_weighted_ws_plev_flux / sum_weighted_ws
    return mass_flux_weighted_pressure


def calc_mass_flux_weighted_latitude(data, lat_min, lat_max, ws_col):
    """
    Calculate mass-flux weighted latitude.

    Component of method from Archer & Caldiera (2008) https://doi.org/10.1029/2008GL033614
    NOTE: Problem with including 1000 hPa

    Parameters
    ----------
    data : xarray.Dataset
        Data containing windspeed and lat
    lat_min : int or float
        Minimum latitude to consider for weighted latitude
    lat_max : int or float
        Maximum latitude to consider for weighted latitude
    ws_col : string
        Name of column to calculate weighted average from (e.g. 'ws', 'ua', 'va')

    Returns
    ----------
    mass_flux_weighted_latitude : xr.DataArray
        Data with mass flux weighted latitude
    """
    assert "lat" in data.coords, "'lat' needs to be in data.coords"

    sub_data = data.sel(lat=slice(lat_min, lat_max))

    sum_weighted_lat_flux = None
    sum_weighted_ws_by_lat = None
    for lat in sub_data["lat"].data:
        lat_data = sub_data.sel(lat=lat, method="nearest")
        lat_sum_weighted_ws = get_mass_weighted_average_wind(lat_data, ws_col)
        if sum_weighted_lat_flux is None:
            sum_weighted_ws_by_lat = lat_sum_weighted_ws
            sum_weighted_lat_flux = lat_sum_weighted_ws * lat
        else:
            sum_weighted_ws_by_lat += lat_sum_weighted_ws
            sum_weighted_lat_flux += lat_sum_weighted_ws * lat
    mass_flux_weighted_latitude = sum_weighted_lat_flux / sum_weighted_ws_by_lat
    return mass_flux_weighted_latitude


def get_climatology(data, freq):
    """
    Makes a climatology at given interval (i.e. days, months, season)

    Parameters
    ----------
    data : xarray.Dataset
        data with regular time stamp
    freq : str
        'day', 'month' or 'season'

    Returns
    ----------
    climatology : xarray.Dataset
        Climatology of a given frequency

    Usage
    ----------
    climatology = get_climatology(data, 'month')

    """
    climatology = data.groupby("time.%s" % (freq)).mean("time")
    return climatology


def calc_low_pass_weights(window, cutoff):
    """
    Calculate weights for a low pass Lanczos filter.
    A low-pass filter removes short-term random fluctations in a time series

    Component of method from Woollings et al (2010) http://dx.doi.org/10.1002/qj.625
    & Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1

    TAKEN FROM:
    https://scitools.org.uk/iris/docs/v1.2/examples/graphics/SOI_filtering.html

    Parameters
    ----------
    window : int
        The length of the filter window.
    cutoff : float
        The cutoff frequency in inverse time steps.

    Returns
    ----------
    output : numpy.array
        filtered weights
    """
    order = ((window - 1) // 2) + 1
    nwts = 2 * order + 1
    w = np.zeros([nwts])
    n = nwts // 2
    w[n] = 2 * cutoff
    k = np.arange(1.0, n)
    sigma = np.sin(np.pi * k / n) * n / (np.pi * k)
    firstfactor = np.sin(2.0 * np.pi * cutoff * k) / (np.pi * k)
    w[n - 1 : 0 : -1] = firstfactor * sigma
    w[n + 1 : -1] = firstfactor * sigma
    return w[0 + (window % 2) : -1]  # edited from w[1:-1]


def apply_lanczos_filter(dataarray, filter_freq, window_size):
    """
    Will carry out Lanczos low-pass filter

    Component of method from Woollings et al (2010) http://dx.doi.org/10.1002/qj.625
    & Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1

    Parameters
    ----------
    datarray : xarray.DataArray
        Data to apply filter to containing zonal mean (u-component) wind
    filter_freq : int
        number of days in filter
    window_size : int
        number of days in window for Lancoz filter

    Returns
    ----------
    window_cons : xarray.DataArray
        Filtered zonal mean data
    """
    if window_size <= filter_freq or filter_freq <= 0 or window_size <= 0:
        raise ValueError(
            "Lanczos filter frequency cannot be larger than window size and both need to be more than 0"
        )

    if not xr.infer_freq(dataarray["time"]) == "D":
        print(
            f"Warning: The 'apply_lanczos_filter' function was built to work on a datetime index of freq 'D'. Frequency in data is '{xr.infer_freq(dataarray['time'])}'"
        )

    assert isinstance(dataarray, xr.DataArray), "Input data needs to be a data array"

    try:
        start_date = dataarray["time"].astype(np.datetime64)[0]
        filter_end_date = start_date + np.timedelta64(
            filter_freq, "D"
        )  # add filter end date for check if data is outside the filter size
        window_end_date = start_date + np.timedelta64(
            window_size, "D"
        )  # add window end date for check if data is outside the filter size
        end_date = dataarray["time"].astype(np.datetime64)[-1]
    except (ValueError, TypeError):
        # makes assumption that if times not coerced into numpy datetime then is 360
        start_date = dataarray["time"].values[0]
        end_date = dataarray["time"].values[-1]
        if not hasattr(start_date, "calendar"):
            raise ValueError(
                f"Error: datetime type inputted ({start_date}) cannot be coerced into numpy.datetime64 and is not in cftime format."
            )
        if start_date.calendar == "360_day":
            datetime_added_func_to_use = data_utils.add_num_of_days_to_360Datetime
        elif start_date.calendar == "noleap":
            datetime_added_func_to_use = data_utils.add_num_of_days_to_NoLeapDatetime
        else:
            raise ValueError(
                f"Error: datetime type inputted ({start_date}) cannot be coerced into numpy.datetime64 and is not in cftime format."
            )
        filter_end_date = datetime_added_func_to_use(
            start_date, num_of_days_to_add=filter_freq
        )
        window_end_date = datetime_added_func_to_use(
            start_date, num_of_days_to_add=window_size
        )

    # add filter day for check if data is outside the filter size
    if window_end_date > end_date or filter_end_date > end_date:
        raise ValueError(
            f"Time series is too short to apply {window_size} day window for Lanczos filter frequency of {filter_freq} days"
        )

    #  As the data may not actually be in day format, it will return the number of values within window in data i.e. if in monthly format but filter window is 60, then actual filter size will be 2
    actual_filter_freq = dataarray.sel(
        time=slice(start_date, filter_end_date)
    ).time.size
    actual_window_size = dataarray.sel(
        time=slice(start_date, window_end_date)
    ).time.size
    lanczos_weights = calc_low_pass_weights(actual_window_size, 1 / actual_filter_freq)
    lanczos_weights_arr = xr.DataArray(lanczos_weights, dims=["window"])
    window_cons = (
        dataarray.rolling(time=len(lanczos_weights_arr), center=True)
        .construct("window")
        .dot(lanczos_weights_arr)
    )
    return window_cons


def get_latitude_and_speed_where_max_ws(data_row):
    """
    Will return the latitude and windspeed at the index of maximum wind speed
    from a row of data
    TODO: expects only one cell to be max

    Component of method from Woollings et al (2010) http://dx.doi.org/10.1002/qj.625
    & Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1
    Barnes & Simpson (2017) https://doi.org/10.1175/JCLI-D-17-0299.1
    & Grise & Polvani (2016) https://doi.org/10.1002/2015JD024687

    Parameters
    ----------
    data_row : xarray.DataArray
        Data of single time unit containing u-component wind

    Returns
    ----------
    lat_at_max : int or float
        Latitude of maximum windspeed
    speed_at_max : int or float
        Speed at latitude of maximum windspeed
    """
    try:
        assert hasattr(data_row, "isnull")
    except Exception as e:
        raise AttributeError("input needs to have isnull method.") from e

    if not data_row.isnull().all():
        data_row = data_row.fillna(0.0)
        max_ws = data_row.where(
            np.abs(data_row) == np.abs(data_row).max(), drop=True
        ).squeeze()
        if max_ws.size > 1:
            print(
                "'get_latitude_and_speed_where_max_ws method': Warning: more than one max value found, picking the first occurence!"
            )
            max_ws = max_ws[0].max()
        lat_at_max = float(max_ws["lat"])
        speed_at_max = float(max_ws.data)
        return lat_at_max, speed_at_max
    else:
        return np.nan, np.nan


def calc_latitude_and_speed_where_max_ws(data, var_col="ua"):
    """
    Will return the latitude and windspeed at the index of maximum wind speed
    from a row of data
    TODO: expects only one cell to be max

    Component of method from Woollings et al (2010) http://dx.doi.org/10.1002/qj.625
    & Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1
    Barnes & Simpson (2017) https://doi.org/10.1175/JCLI-D-17-0299.1
    & Grise & Polvani (2016) https://doi.org/10.1002/2015JD024687

    Parameters
    ----------
    dataarray : xarray.dataset
        Dataset containing u-component wind

    Returns
    ----------
    jet_lats :
        Latitude of maximum windspeed
    jet_speeds : int or float
        Speed at latitude of maximum windspeed
    """
    inds = data[var_col].argmax(dim="lat", skipna=True)
    jet_lats = data[var_col]["lat"][inds]
    data["jet_lat"] = ((inds.dims), jet_lats.data)
    data["jet_speed"] = data[var_col].max(dim="lat")
    # cleans up missing values
    data = data.dropna("time", subset=["jet_lat", "jet_speed"])
    return data


def assign_jet_lat_and_speed_to_data(
    data, max_lat_ws, max_lats_col="jet_lat", max_ws_col="jet_speed"
):
    """
    Will return a data array with the maximum windspeed and latitude of that maximum wind speed

    Component of method from Woollings et al (2010) http://dx.doi.org/10.1002/qj.625
    & Barnes & Simpson 2017 https://doi.org/10.1175/JCLI-D-17-0299.1

    Parameters
    ----------
    data : xarray.DataArray
        Data of single time unit containing u-component wind
    max_lat_ws : np.array
        Array of maximum latitude and windspeed at those maximums for each timeunit

    Returns
    ----------
    data_with_max_lats_ws : xarray.Dataset
        Data with maximum latitude and windspeed attached
    """
    max_lats = max_lat_ws[:, 0]
    max_ws = max_lat_ws[:, 1]
    data_with_max_lats_ws = data.assign(
        {max_lats_col: (("time"), max_lats), max_ws_col: (("time"), max_ws)}
    )
    data_with_max_lats_ws[max_lats_col] = data_with_max_lats_ws[max_lats_col].astype(
        float
    )
    data_with_max_lats_ws[max_ws_col] = data_with_max_lats_ws[max_ws_col].astype(float)
    return data_with_max_lats_ws


def apply_low_freq_fourier_filter(data, highest_freq_to_keep):
    """
    Carries out a Fourier transform for filtering keeping only low frequencies.

    Component of method from Woollings et al (2010) http://dx.doi.org/10.1002/qj.625

    ADAPTED FROM: https://scipy-lectures.org/intro/scipy/auto_examples/plot_fftpack.html

    Parameters
    ----------
    data : np.array - 1-d
        time series data at regular intervals
    highest_freq_to_keep : int
        highest frequency to keep in the fourier transform expression
        NOTE: starts at 0, so highest_freq_to_keep=1 will only keep the constant and first expresion

    Returns
    ----------
    filtered_sig : np.array
        Fourier-filtered signal

    Usage
    ----------
    # Apply filter of the two lowest frequencies
    apply_low_freq_fourier_filter(data, highest_freq_to_keep=2)
    """
    # Fast Fourier Transform on the time series data
    fourier_transform = scipy.fftpack.fft(data)

    # Remove low frequencies
    fourier_transform[highest_freq_to_keep + 1 :] = 0

    # Inverse Fast Fourier Transform the time series data back
    filtered_sig = scipy.fftpack.ifft(fourier_transform)
    return filtered_sig


def assign_filtered_lats_and_ws_to_data(data, filtered_max_lats, filtered_max_ws, dim):
    """
    Assigns the filtered data back to the returned dataset

    Component of method from Woollings et al (2010) http://dx.doi.org/10.1002/qj.625

    Parameters
    ----------
    data : xarray.Dataset
        Data to have fourier filtered data assigned to
    filtered_max_lats : numpy.array
        Fourier-filtered maximum latitude by given timeunit
    filtered_max_lats : numpy.array
        Fourier-filtered maximum speed at maximum latitude by given timeunit

    Returns
    ----------
    filtered_data : xarray.Dataset
        Data with fourier-filtered maximum lat and windspeed attached to it
    """
    filtered_data = data.assign(
        {
            "ff_jet_lat": ((dim), filtered_max_lats),
            "ff_jet_speed": ((dim), filtered_max_ws),
        }
    )
    filtered_data["ff_jet_lat"] = filtered_data["ff_jet_lat"].astype(float)
    filtered_data["ff_jet_speed"] = filtered_data["ff_jet_speed"].astype(float)
    return filtered_data


def calc_jet_width_for_one_day(data_row, jet_lat, jet_speed):
    """
    Calculates jet width on a given data array using the calculated jet lat and speed at that lat

    Component of method from Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1

    Parameters
    ----------
    data_row : xarray.DataArray
        Data of single time unit containing u-component wind
    jet_lat : int or float
        Latitude of jet
    jet_speed : int or float
        Speed of jet at jet_lat

    Returns
    ----------
    jet_widths : array-like
        Width of jet-stream in latitude. Units for latitude are the same as the input units i.e. likely to be 'degrees_north'
    """
    lat_resolution = float(data_row["lat"][1] - data_row["lat"][0])
    if not jet_speed:
        return np.nan
    possible_surrounding_jet_vals = get_possible_surrounding_jet_lat_vals_for_one_day(
        data_row, jet_speed
    )
    if possible_surrounding_jet_vals.size == data_row["lat"].size:
        print("No jet-width determined")
        return np.nan
    possible_surrounding_jet_lat_vals = possible_surrounding_jet_vals["lat"]
    jet_width = get_width_of_jet_from_surrounding_lat_vals(
        jet_lat, possible_surrounding_jet_lat_vals, lat_resolution
    )
    return jet_width


def get_possible_surrounding_jet_lat_vals_for_one_day(one_day_wind_data, jet_speed):
    """
    Returns array of windspeed values that are within half speed of the jet_speed given.

    Component of method from Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1

    Parameters
    ----------
    one_day_wind_data : xarray.DataArray
        Data of single time unit containing u-component wind
    jet_speed : int or float
        Speed of jet at jet_lat

    Returns
    ----------
    possible_surrounding_jet_vals : array-like
        DataArray of True or False as to whether the windspeed value is within half speed of the jet_speed
    """
    half_jet_speed = jet_speed / 2
    possible_surrounding_jet_vals = one_day_wind_data[
        one_day_wind_data > half_jet_speed
    ]
    return possible_surrounding_jet_vals


def get_width_of_jet_from_surrounding_lat_vals(
    jet_lat, possible_surrounding_jet_lat_vals, lat_resolution
):
    """
    Get latitude width of jet around the given jet latitude value. Width given in units of latitude

    Component of method from Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1

    Parameters
    ----------
    jet_lat : int or float
        Latitude of jet
    possible_surrounding_jet_lat_vals : array-like
        DataArray of lats where the windspeed value is within half speed of the jet_speed
    lat_resolution : int or float
        Latitude resolution in degrees

    Returns
    ----------
    jet_width : array-like
        Width of jet-stream in latitude. Unit for latitude are the same as the input units i.e. likely to be 'degrees_north'
    """
    lats_within_jet = get_all_surrounding_jet_lats_around_max_lat(
        jet_lat, possible_surrounding_jet_lat_vals, lat_resolution
    )
    if not lats_within_jet:
        return np.nan
    jet_width = max(lats_within_jet) - min(lats_within_jet)
    return jet_width


def get_all_surrounding_jet_lats_around_max_lat(
    jet_lat, possible_surrounding_jet_lat_vals, lat_resolution
):
    """
    Returns all the latitudes in a continuous group (i.e. within one unit of lat resolution) around selected jet (jet_lat).

    Component of method from Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1

    Parameters
    ----------
    jet_lat : int or float
        Latitude of jet
    possible_surrounding_jet_lat_vals : array-like
        DataArray of lats where the windspeed value is within half speed of the jet_speed
    lat_resolution : int or float
        Latitude resolution in degrees

    Returns
    ----------
    lats_within_jet : list
        All latitudes in continous group (i.e. within lat_resolution) around jet_lat
    """
    lats_within_jet = []
    lats_at_higher_lats = check_lats_within_jet_in_one_direction(
        jet_lat, possible_surrounding_jet_lat_vals, lat_resolution
    )
    lats_at_lower_lats = check_lats_within_jet_in_one_direction(
        jet_lat, possible_surrounding_jet_lat_vals, -lat_resolution
    )
    lats_within_jet.extend(lats_at_higher_lats)
    lats_within_jet.extend(lats_at_lower_lats)
    return lats_within_jet


def check_lats_within_jet_in_one_direction(
    jet_lat, possible_surrounding_jet_lat_vals, amount_to_add_to_lat_val
):
    """
    Creates a list of latitudes that are within a given distance from the jet_lat and only in one direction i.e. increasing or decreasing

    Component of method from Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1

    Parameters
    ----------
    jet_lat : int or float
        Latitude of jet
    possible_surrounding_jet_lat_vals : array-like
        DataArray of lats where the windspeed value is within half speed of the jet_speed
    amount_to_add_to_lat_val : int or float
        Check if lat within this amount

    Returns
    ----------
    lats_within_jet : list
        All latitudes in continous group (i.e. within lat_resolution) around jet_lat
    """
    lats_within_jet = []
    current_lat = jet_lat
    current_lat_is_within_range = True
    while current_lat_is_within_range:
        if current_lat in possible_surrounding_jet_lat_vals:
            lats_within_jet.append(current_lat)
        else:
            current_lat_is_within_range = False
        current_lat += amount_to_add_to_lat_val
    return lats_within_jet


def scale_lat_vals_with_quadratic_func(lats, speeds, scaled_lats):
    """
    Will downscale or upscale the resolution of latitude using a quadratic func.

    Component of method from Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1
    & Grise & Polvani (2016) https://doi.org/10.1002/2015JD024687

    Parameters
    ----------
    lats : xr.DataArray or array-like
        Array of latitude values
    speeds :  xr.DataArray or array-like
        Array of wind-speeds
    scaled_lats : array-like
        Array of scaled latitude values of a given resolution (see rescale_lat_resolution function)

    Returns
    ----------
    scaled_lat_vals : xr.DataArray or array-like
        Array of rescaled latitude values based scaled_lats
    """
    scaled_lat_vals = apply_quadratic_func(lats, speeds, scaled_lats)
    return scaled_lat_vals


def get_latitude_and_speed_where_max_ws_at_reduced_resolution(
    lats_and_ws, lat_resolution
):
    """
    Get latitude and speed where max windspeed at a reduced resolution.
    Makes use of the quadratic func to scale latitude values.

    Component of method from Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1
    & Grise & Polvani (2016) https://doi.org/10.1002/2015JD024687

    Parameters
    ----------
    lats_and_ws : xr.DataArray or array-like
        Array of latitudes and windspeeds
    lat_resolution : int or float
        Latitude resolution in degrees

    Returns
    ----------
    output : numpy.array
        latitude and wind-speed value scaled by quadratic func
    """
    lats, ws = lats_and_ws
    #  Remove numpy.nan from list
    lats = [lat for lat in lats if not np.isnan(lat)]
    ws = [s for s in ws if not np.isnan(s)]
    #  Scale lats
    scaled_lats = data_utils.rescale_lat_resolution(lats, lat_resolution)
    scaled_lat_vals = scale_lat_vals_with_quadratic_func(lats, ws, scaled_lats)
    decimal_places = data_utils.get_num_of_decimal_places(lat_resolution)
    max_speed_at_scaled_lat = np.max(scaled_lat_vals)
    return (
        round(scaled_lats[np.argmax(scaled_lat_vals)], decimal_places),
        max_speed_at_scaled_lat,
    )


def get_3_latitudes_and_speed_around_max_ws(row):
    """
    Will get the latitudes neighbouring to east, at and to west where the max windspeed is found

    Component of method from Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1
    & Grise & Polvani (2016) https://doi.org/10.1002/2015JD024687

    Parameters
    --------------
    row : xr.DataArray or array-like
        Array containing latitude and wind-speed values

    Returns
    ----------
    neighbouring_lats : array-like
        array of 3 neighbouring latitude coordinates around where maximum windspeed is found in input (row)
    neighbouring_speeds : array-like
        array of 3 neighbouring winspeeds values around where maximum windspeed is found in input (row)
    """
    assert "lat" in row.coords, "'lat' needs to be in data.coords"

    max_lat, _ = get_latitude_and_speed_where_max_ws(row)
    if np.isnan(max_lat):
        # occurs when no data in slice
        return (
            np.array([np.nan, np.nan, np.nan], dtype="float64"),
            np.array([np.nan, np.nan, np.nan], dtype="float64"),
        )
    max_lat_argwhere = int(np.argwhere(row["lat"].data == max_lat))

    if max_lat_argwhere == 0:
        neighbouring_lats_ind = [max_lat_argwhere, max_lat_argwhere + 1]
    elif max_lat_argwhere == row["lat"].data.size - 1:
        neighbouring_lats_ind = [max_lat_argwhere - 1, max_lat_argwhere]
    else:
        neighbouring_lats_ind = [
            max_lat_argwhere - 1,
            max_lat_argwhere,
            max_lat_argwhere + 1,
        ]
    neighbouring_lats = row["lat"].isel(lat=neighbouring_lats_ind).data
    neighbouring_speeds = row.isel(lat=neighbouring_lats_ind).data

    #  add nan value to edges
    if len(neighbouring_lats) < 3:
        neighbouring_lats = add_nan_value_to_arr_if_not_len_enough(
            neighbouring_lats, len_required=3
        )
        neighbouring_speeds = add_nan_value_to_arr_if_not_len_enough(
            neighbouring_speeds, len_required=3
        )
    return (neighbouring_lats, neighbouring_speeds)


def add_nan_value_to_arr_if_not_len_enough(arr, len_required):
    """
    Add NaN value to array if any values in the array so it is a certain length

    Component of method from Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1
    & Grise & Polvani (2016) https://doi.org/10.1002/2015JD024687

    Parameters
    ----------
    arr : array-like
        Array to check.
    len_required : int
        length of array required

    Returns
    ----------
    arr : array-like
        new array

    """
    assert (
        len(arr) <= len_required
    ), "length of input array needs to be less or same as len_required"
    if len(arr) != len_required:
        arr = np.append(arr, np.nan)
        return add_nan_value_to_arr_if_not_len_enough(arr, len_required)
    return arr


def get_3_neighbouring_coord_values(coord_val, coord_resolution):
    """
    Get the three neighbouring values to a given input coord (coord_val)

    Component of method from Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1
    & Grise & Polvani (2016) https://doi.org/10.1002/2015JD024687

    Parameters
    ----------
    coord_val : int or float
        Central coord value to get neighbours from
    coord_resolution : int or float
        in degrees

    Returns
    ----------
    output : array-like
        array of 3 neighbouring latitude coordinates

    Usage
    ----------
    get_3_neighbouring_coord_values(45.0, 1.25)
    >>> np.array([43.75, 45.0, 46.25])
    """
    if not isinstance(coord_val, float) or not isinstance(coord_resolution, float):
        coord_val = float(coord_val)
        coord_resolution = float(coord_resolution)

    return np.array(
        [coord_val - coord_resolution, coord_val, coord_val + coord_resolution]
    )


def quadratic_func(x, y):
    """
    Quadratic function.

    Component of method from Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1
    & Grise & Polvani (2016) https://doi.org/10.1002/2015JD024687

    Parameters
    ----------
    x : xr.DataArray or array-like
        Array 1
    y : xr.DataArray or array-like
        Array 2
    """
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=np.RankWarning)
        p = np.polyfit(x, y, deg=2)
    return p


def apply_quadratic_func(x, y, vals):
    """
    Apply quadratic function to an array of values (vals).

    Component of method from Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1
    & Grise & Polvani (2016) https://doi.org/10.1002/2015JD024687


    Parameters
    ----------
    x : xr.DataArray or array-like
        Array 1
    y : xr.DataArray or array-like
        Array 2
    vals : array-like
        Values to apply function to

    Returns
    ----------
    output : array-like
        quadratic function output
    """
    a, b, c = quadratic_func(x, y)
    return (a * vals**2) + (b * vals) + c


def get_jet_lat_and_speed_using_parabola_by_day(data_row):
    """
    Will get jet latitude and speed by fitting a parabola and taking maximum

    Component of method from Barnes & Polvani (2015) http://journals.ametsoc.org/doi/10.1175/JCLI-D-14-00589.1

    Parameters
    ----------
    data_row : xarray.DataArray
        Data of single time unit containing u-component wind

    Returns
    ----------
    data_row : xarray.DataArray
        Data of single time unit containing jet_lat and jet_speed variables
    """

    data_ua = abs(data_row["ua"].dropna(dim="lat"))
    try:
        # Try and fit the parabola
        _, coeff = fit_parabola(data_ua["lat"].data, data_ua.data)
    except (RuntimeError, ValueError):
        coeff = [np.nan, np.nan]
    # Check that the max lat is not above the min or max (problem with using a parabola)
    if coeff[0] < data_row["lat"].min() or coeff[0] > data_row["lat"].max():
        coeff = [np.nan, np.nan]
    data_row["jet_lat"] = float(coeff[0])
    data_row["jet_speed"] = float(coeff[1])
    return data_row


def fit_parabola(x, y):
    """
    Fits a parabola.

    Component of method from Barnes & Polvani (2015) http://journals.ametsoc.org/doi/10.1175/JCLI-D-14-00589.1
    """
    coeff, _ = scipy.optimize.curve_fit(parabola, x, y)
    fitted_parabola = parabola(x, coeff[0], coeff[1], coeff[2])
    return fitted_parabola, coeff


def parabola(x, lat, speed, a):
    """
    Parabola.

    Component of method from Barnes & Polvani (2015) http://journals.ametsoc.org/doi/10.1175/JCLI-D-14-00589.1
    """
    return -a * (x - lat) ** 2 + speed


def assign_jet_lat_speed_to_ten_day_mean_data(ten_day_mean):
    """
    Joins the values for latitude and windspeed of the point of maximum windspeed for a given sector to the ten_day_mean data.

    Component of method from Barnes & Simpson 2017 https://doi.org/10.1175/JCLI-D-17-0299.1

    Parameters
    ----------
    ten_day_mean : xarray.Dataset
        Data containing u-component wind and resampled by 10 days

    Returns
    ----------
    ten_day_mean : xarray.Dataset
        Data with jet latitude and windspeed
    """
    max_lats_ws = np.array(
        list(map(get_latitude_and_speed_where_max_ws, ten_day_mean["ua"]))
    )
    ten_day_mean = assign_jet_lat_and_speed_to_data(ten_day_mean, max_lats_ws)
    return ten_day_mean


def assign_ten_day_average_jet_lat_speed_to_data(data, ten_day_mean):
    """
    Joins the values for latitude and windspeed of the point of maximum windspeed to the main data

    Component of method from Barnes & Simpson 2017 https://doi.org/10.1175/JCLI-D-17-0299.1

    Parameters
    ----------
    data : xarray.Dataset
        Data containing u-component wind

    Returns
    ----------
    output : xarray.Dataset
        Data with jet latitude and windspeed
    """
    return data.assign(
        {
            "jet_lat": (("10_day_average"), ten_day_mean["jet_lat"].data),
            "jet_speed": (("10_day_average"), ten_day_mean["jet_speed"].data),
        }
    )


def cubic_spline_interpolation(x, y):
    """
    Runs a cubic spline interpolation using 2 equal-length arrays.

    Component of method from Bracegirdle et al (2018) https://doi.org/10.1175/JCLI-D-17-0320.1

    Parameters
    ----------
    x : xarray.DataArray or array-like
        array 1 to run cubic spline interpolation
    y : xarray.DataArray or array-like
        array 2 to run cubic spline interpolation

    Returns
    ----------
    cubic_interpolation : np.array
        Cubic spline interpolation
    """
    return scipy.interpolate.interp1d(x, y, kind="cubic", fill_value="extrapolate")


def run_cubic_spline_interpolation_to_get_max_lat_and_ws(
    data, lat_resolution, ua_col="ua"
):
    """
    Runs a cubic spline interpolation to find maximum latitude and maximum windspeed at a given resolution of latitude.

    Component of method from Bracegirdle et al (2018) https://doi.org/10.1175/JCLI-D-17-0320.1

    Parameters
    ----------
    data : xr.Dataset
        must contain coords lat
    lat_resolution : int or float
        Latitude resolution in degrees for cubic spline interpolation
    ua_col : str
        u-component windspeed column (default='ua')

    Returns
    ----------
    max_lat : float
        latitude with the maximum wind-speed for time period as deterimined by cubic spline interpolation to a given lat resolution
    max_ws : float
        maximum wind-speed at the given latitude for time period
    """
    scaled_lats = data_utils.rescale_lat_resolution(data["lat"], lat_resolution)
    csi = cubic_spline_interpolation(data["lat"], data[ua_col])
    interpolated_ws = csi(scaled_lats)
    max_lat = scaled_lats[np.argmax(interpolated_ws)]
    max_ws = max(interpolated_ws)
    return max_lat, max_ws


def run_cubic_spline_interpolation_for_each_unit_of_climatology_to_get_max_lat_and_ws(
    data, lat_resolution, time_col
):
    """
    Runs a cubic spline interpolation to find maximum latitude and maximum windspeed at a given resolution of latitude for each.

    Component of method from Bracegirdle et al (2018) https://doi.org/10.1175/JCLI-D-17-0320.1

    Parameters
    ----------
    data : xarray.Dataset
        Data containing u-component wind-speed. In the case of Bracegirdle et al. 2018, uses a seasonal and annual mean/climatology
    resolution : int or float
        Latitude resolution in degrees for cubic spline interpolation
    time_col : str
        Column name containing period (i.e. season, year, month, etc.)

    Returns
    ----------
    max_lats : list
        list of latitudes with the maximum wind-speed
    max_ws : list
        list of maximum wind-speed at the given latitudes

    Usage
    ----------
    seasonal_climatology = jetstream_metrics_components.get_climatology(data, "season")
    seasonal_zonal_mean = seasonal_climatology.mean("lon")
    (
        seasonal_max_lats,
        seasonal_max_ws,
    ) = jetstream_metrics_utils.run_cubic_spline_interpolation_for_each_unit_of_climatology_to_get_max_lat_and_ws(
        seasonal_zonal_mean, resolution=0.075, time_col="season"
    )
    """
    max_lats = []
    max_ws = []
    for period in data[time_col].data:
        period_data = data.sel({time_col: period})
        lat, ws = run_cubic_spline_interpolation_to_get_max_lat_and_ws(
            period_data, lat_resolution=lat_resolution
        )
        max_lats.append(lat)
        max_ws.append(ws)
    return max_lats, max_ws


def calc_centroid_jet_lat_from_zonal_mean(zonal_mean, area_by_lat):
    """
    Will get the centroid latitude of the u-component wind using:
    jet_lat = integral(60deg, 30deg)(zonal_u**2*lat) dlat / integral(60deg, 30deg)(zonal_u**2) dlat

    Component of method from Ceppi et al (2018) https://doi.org/10.1175/JCLI-D-17-0323.1 & Zappa et al. 2018 https://doi.org/10.1029/2019GL083653

    Parameters
    ----------
    zonal_mean : xarray.Dataset
        zonally-average data containing u-component wind data
    area_by_lat : xarray.DataArray
        Information on area in each latitude (i.e. m2 per latitude)
    """
    u_hat_by_lat = zonal_mean["ua"] ** 2 * zonal_mean["lat"] * area_by_lat
    u_hat = zonal_mean["ua"] ** 2 * area_by_lat
    return u_hat_by_lat.sum("lat") / u_hat.sum("lat")


def get_latitude_value_in_data_row(data_row, lat_val):
    """
    Will return the latitude value of a given data row. Needed to loop through data and external array

    Component of method from Ceppi et al (2018) https://doi.org/10.1175/JCLI-D-17-0323.1 & Zappa et al. 2018 https://doi.org/10.1029/2019GL083653

    Parameters
    ----------
    data_row : xarray.DataArray
        Data row to extract variable from using given latitude
    lat_val : float or int
        latitude value to extract from data row
    """
    return float(data_row.sel(lat=lat_val, method="nearest"))


def get_moving_averaged_smoothed_jet_lats_for_one_day(data_row, width_of_pulse=10):
    """
    Get moving average of smoothed_jet_lats for one day. Used in combination with a groupby mapping.

    Component of method from Kerr et al. (2020) https://onlinelibrary.wiley.com/doi/10.1029/2020JD032735

    Parameters
    ----------
    data_row : xarray.DataArray
        Data containing u-component windspeed of one time unit

    Returns
    ----------
    smoothed_jet_lat : xarray.Dataset
        Data containing jet-stream position
    """
    data_row["jet_lat"] = get_jet_lat_by_lon(data_row["ua"])
    data_row["smoothed_jet_lat"] = smooth_jet_lat_across_lon_with_rectangular_pulse(
        data_row["jet_lat"], width_of_pulse=width_of_pulse
    )
    return data_row


def smooth_jet_lat_across_lon_with_rectangular_pulse(jet_lat_data, width_of_pulse):
    """
    Smooth jet position (jet latitude) by carrying out a convolution with a rectangular pulse.

    Component of method from Kerr et al. (2020) https://onlinelibrary.wiley.com/doi/10.1029/2020JD032735

    Parameters
    ----------
    jet_lat_data : xarray.DataArray
        Data detailing the latitude of jet-stream at each longitude of one time unit
    width_of_pulse : float or int
        Width to make rectangular pulse (likely to be in  'degrees_east')

    Returns
    ----------
    filtered_data : xarray.DataArray
        Data detailing the latitude of jet-stream that has been smoothed using rectangular pulse

    """
    sig = jet_lat_data["lon"] % jet_lat_data < width_of_pulse
    filtered_data = jet_lat_data[sig]
    return filtered_data


def get_jet_lat_by_lon(data_row):
    """
    Gets all the latitudes of maximum wind-speeds by longitude.

    Component of method from Kerr et al. (2020) https://onlinelibrary.wiley.com/doi/10.1029/2020JD032735

    Parameters
    ----------
    data_row : xarray.DataArray
        Data containing u-component windspeed of one time unit

    Returns
    ----------
    output : xarray.DataArray
        Data detailing the latitude of jet-stream at each longitude of one time unit
    """
    return data_row["lat"][data_row.argmax("lat")]



# -*- coding: utf-8 -*-

"""
    Statistics for isolating individual quantities synonymous with the jet stream from upper-level wind speed
    within a given time window (e.g. latitude, speed, width).

    The following statistics each return a xarray.Dataset and are ordered by paper publish year.
"""

# imports
import numpy as np
import xarray
from . import jet_statistics_components
from jsmetrics.utils import data_utils, spatial_utils, windspeed_utils
from jsmetrics.core.check_data import sort_xarray_data_coords

# docs
__author__ = "Thomas Keel"
__email__ = "thomas.keel.18@ucl.ac.uk"
__status__ = "Development"


@sort_xarray_data_coords(coords=["lat", "lon"])
def archer_caldeira_2008(data):
    r"""
    This method calculates three monthly-averaged jet stream properties (windspeed, pressure and latitude) via integrated quantities from u-component wind.

    This method returns three properties:

    1. **weighted-average wind speed** -- jet stream wind speed (:math:`WS`), calculated by:

    .. math::
        WS_{i, j} =  \frac{\sum\limits_{k=400hPa}^{k=100hPa} m_{k} \times \sqrt{u^{2}_{i, j, k} + v^{2}_{i, j, k}}}{\sum\limits_{k=400hPa}^{k=100hPa} m_{k}}

    where :math:`u_{i,j,k}` and :math:`v_{i,j,k}` are the monthly-average horizontal wind components at grid point (i,j,k), and :math:`m_{k}` is the mass at level `k`.

    2. **mass flux weighted pressure** -- the average pressure of flows by the tropopause (:math:`P`), calculated by:

    .. math::
        P_{i, j} =  \frac{\sum\limits_{k=400hPa}^{k=100hPa} \left(m_{k} \times \sqrt{u^{2}_{i, j, k} + v^{2}_{i, j, k}}\right) \times p_k}{\sum\limits_{k=400hPa}^{k=100hPa} m_{k} \times \sqrt{u^{2}_{i, j, k} + v^{2}_{i, j, k}}}

    where :math:`p_k` is the pressure at level :math:`k`.

    3. **mass flux weighted latitude** -- Latitude of the Northern Hemisphere jet (:math:`L^{NH}`), calculated by:

    .. math::
        L_{i}^{NH} =  \frac{\sum\limits_{j=15°N}^{j=70°N} \left[\sum\limits_{k=400hPa}^{k=100hPa} \left(m_{k} \times \sqrt{u^{2}_{i, j, k} + v^{2}_{i, j, k}}\right) \right] \times \phi_{i,j}}{\sum\limits_{j=15N}^{j=70N} \sum\limits_{k=400hPa}^{k=100hPa} m_{k} \times \sqrt{u^{2}_{i, j, k} + v^{2}_{i, j, k}}}

    where :math:`\phi_{i,j}` is the grid cell latitude.

    This method was originally introduce in Archer & Caldiera (2008) (https://doi.org/10.1029/2008GL033614)
    and is described in Section 3 of that study.

    **Note:** this method does not explicitly limit inputted wind to 100-400 hPa, see 'Notes' for more information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua' and 'va', and the coordinates: 'lon', 'lat', 'plev' and 'time'.

    Returns
    ----------
    output : xarray.Dataset
        Data containing the three jet properties: 'mass_weighted_average_ws', 'mass_flux_weighted_pressure' and 'mass_flux_weighted_latitude'

    Notes
    -----
    While the initial methodology provides limits for pressure level (100-400 hPa), here the mass weighted outputs
    will be calculated for any pressure levels passed into the method.

    The latitude calculation is limited to 15-70N (as we only provide a way to extract the Northern Hemisphere jet),
    but you may find it easy enough to edit this method to calculate outputs for a different region.

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u and v components:
        uv_data = xr.open_dataset('path_to_uv_data')

        # Subset dataset to range used in original methodology for the NH jet (100-400 hPa & 15-70 N)):
        uv_sub = uv_data.sel(plev=slice(100, 400), lat=slice(15, 70))

        # Run statistic:
        archer_outputs = jsmetrics.jet_statistics.archer_caldiera_2008(uv_sub)

        # Subset mass weighted wind by a windspeed threshold
        windspeed_threshold = 15 # remember this is for monthly averages
        strong_jet = archer_outputs['mass_weighted_average_ws'].where(lambda row: row > windspeed_threshold)

    """
    #  Step 1. Calculate monthly means
    if "time" not in data.coords:
        raise KeyError("Please provide a time coordinate for data to run this metric")
    if data["time"].size == 1:
        print(
            "Warning: only found one time step, and the time coord is being renamed month."
        )
        mon_mean = data.assign_coords(month=data["time"].dt.month)
        mon_mean = mon_mean.expand_dims("month")
    else:
        mon_mean = data.groupby("time.month").mean()

    #  Step 2. Calculate wind-speed from u and v-component wind
    mon_mean["ws"] = windspeed_utils.get_resultant_wind(mon_mean["ua"], mon_mean["va"])

    #  Step 3. Calculate mass weighted average windspeed
    mass_weighted_average = jet_statistics_components.calc_mass_weighted_average(
        mon_mean, ws_col="ws"
    )

    #  Step 4. Calculate mass flux weighted pressure
    mass_flux_weighted_pressure = (
        jet_statistics_components.calc_mass_flux_weighted_pressure(
            mon_mean, ws_col="ws"
        )
    )

    #  Step 5. Calculate mass flux weighted latitude
    mass_flux_weighted_latitude = (
        jet_statistics_components.calc_mass_flux_weighted_latitude(
            mon_mean, lat_min=15, lat_max=75, ws_col="ws"
        )
    )

    #  Step 6. Assign the three new output variables to the original data
    output = data.assign(
        {
            "mass_weighted_average_ws": (
                ("month", "lat", "lon"),
                mass_weighted_average.data,
            ),
            "mass_flux_weighted_pressure": (
                ("month", "lat", "lon"),
                mass_flux_weighted_pressure.data,
            ),
            "mass_flux_weighted_latitude": (
                ("month", "lon"),
                mass_flux_weighted_latitude.data,
            ),
        }
    )
    return output


@sort_xarray_data_coords(coords=["lat", "lon"])
def woollings_et_al_2010(data, filter_freq=10, window_size=61):
    r"""
    This method follows an in-text description of 4-steps describing the algorithm of jet-stream identification
    from Woollings et al. (2010).

    This method returns four outputs:
        1. **jet_lat** -- latitude of maximum speed within low-pass filtered zonally averaged wind profile
        2. **jet_speed** -- speed at the 'jet_lat'
        3. **ff_jet_lat** -- Fourier-filtered 'jet_lat' by season
        4. **ff_jet_speed** -- Fourier-filtered 'jet_speed' by season

    This method was first introduce in Woollings et al (2010) (http://dx.doi.org/10.1002/qj.625) and
    is described in section 2 of that study.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package including Step 5 of the methodology.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua', and the coordinates: 'lon', 'lat', 'plev' and 'time'.
    filter_freq : int
        number of days in filter (default=10 days)
    window_size : int
        number of days in window for Lancoz filter (default=61 days)

    Returns
    ----------
    output : xarray.Dataset
        Data containing the four output variables: 'jet_lat', 'jet_speed', 'ff_jet_lat' and 'ff_jet_speed'

    Notes
    -----
    In the original paper, a further step (Step 5) is carried out to express the values of jet latitude
    and jet speed anomalies from the seasonal cycle, this is shown in the 'Examples'

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u component wind:
        ua_data = xr.open_dataset('path_to_u_data')

        # Subset dataset to range used in original methodology (700-925 hPa & 20-70 N, 300-360 W)):
        ua_sub = ua.sel(plev=slice(700, 925), lat=slice(20, 70), lon=slice(300, 360))

        # Run statistic with a filter frequency and window size used in the original methodology:
        w10 = jsmetrics.jet_statistics.woollings_et_al_2010(ua_sub, filter_freq=10, window_size=61)

        # Express jet latitude and speed as anomalies from smoothed seasonal cycle (Step 5 of methodology)
        w10_seasonal_anomalies = w10.groupby('time.season').apply(lambda row: row['jet_lat'] - row['ff_jet_lat'])

    """
    # Initial translation of data from dataArray to dataset
    if isinstance(data, xarray.DataArray):
        data = data.to_dataset()

    # Step 1: Calculate long and/or plev mean
    zonal_mean = windspeed_utils.get_zonal_mean(data)

    # Step 2: Apply n-day lancoz filter
    lancoz_filtered_mean_data = jet_statistics_components.apply_lanczos_filter(
        zonal_mean["ua"], filter_freq, window_size
    )

    # Step 3: Calculate max windspeed and lat where max ws found
    max_lat_ws = np.array(
        list(
            map(
                jet_statistics_components.get_latitude_and_speed_where_max_ws,
                lancoz_filtered_mean_data[:],
            )
        )
    )
    zonal_mean_lat_ws = jet_statistics_components.assign_jet_lat_and_speed_to_data(
        zonal_mean, max_lat_ws
    )
    # Step 4: Make seasonal climatology
    climatology = jet_statistics_components.get_climatology(zonal_mean_lat_ws, "season")

    # Step 5: Apply low-freq fourier filter to both max lats and max ws
    fourier_filtered_lats = jet_statistics_components.apply_low_freq_fourier_filter(
        climatology["jet_lat"].values, highest_freq_to_keep=2
    )
    fourier_filtered_ws = jet_statistics_components.apply_low_freq_fourier_filter(
        climatology["jet_speed"].values, highest_freq_to_keep=2
    )

    #  Step 6. Assign the new output variables to the original data
    time_dim = climatology["jet_speed"].dims[0]
    output = jet_statistics_components.assign_filtered_lats_and_ws_to_data(
        zonal_mean_lat_ws,
        fourier_filtered_lats.real,
        fourier_filtered_ws.real,
        dim=time_dim,
    )

    return output


@sort_xarray_data_coords(coords=["lat", "lon"])
def barnes_polvani_2013(data, filter_freq=10, window_size=41):
    r"""
    This method constructs the 'eddy-driven jet' by performing a pressure-weighted average of zonal winds. The winds
    are then low-pass frequency filtered at each grid point using a 10-day Lanczos filter with 41 weights by default.
    Finally a 0.01 degree quadratic function is fitted to the peak of the subsequent wind speed profile for each time step.

    This method returns three outputs:
        1. **jet_lat** -- latitude of maximum speed at an interval of 0.01 degree within the subseqeunt profile
        2. **jet_speed** -- speed at the 'jet_lat'
        3. **jet_width** -- full width at half of the maximum 'jet_speed' within the 0.01 quadratic fitted to the peak of the wind speed profile

    *Note:* 'jet_width' is undefined where the 'jet_speed' never drops below half maximum.

    This method was originally introduce in Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1
    and is described in Section 2b and 2c of that study.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua', and the coordinates: 'lon', 'lat', 'plev' and 'time'.
    filter_freq : int
        number of days in filter (default=10 timeunits)
    window_size : int
        number of days in window for low-pass Lancoz filter (default=41 timeunits)

    Returns
    ----------
    output : xarray.Dataset
        Data containing the three output variables: 'jet_lat', 'jet_speed', 'jet_width'

    Notes
    -----
    Whereas the original analysis using this method focuses on three distinct sections of the globe,
    the method here does not make any distinction. Instead we highlight how to do subset the input data and
    calculate this metric for those three sections in 'Examples.

    This method was based on the method from Woollings et al. (2010) (http://dx.doi.org/10.1002/qj.625)

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u component wind:
        ua_data = xr.open_dataset('path_to_u_data')

        # Subset dataset to three sectors of the globe used in original methodology:
        ua_sh = ua.sel(plev=slice(700, 850), lat=slice(-90, 0)) # the Southern Hemisphere
        ua_na = ua.sel(plev=slice(700, 850), lat=slice(0, 90), lon=slice(300, 360) # the North Atlantic
        ua_np = ua.sel(plev=slice(700, 850), lat=slice(0, 90), lon=slice(135, 235) # the North Pacific

        # Run statistic with a filter frequency and window size used in the original methodology:
        bp13_sh = jsmetrics.jet_statistics.barnes_polvani_2013(ua_sh, filter_freq=10, window_size=41)
        bp13_na = jsmetrics.jet_statistics.barnes_polvani_2013(ua_na, filter_freq=10, window_size=41)
        bp13_np = jsmetrics.jet_statistics.barnes_polvani_2013(ua_np, filter_freq=10, window_size=41)
    """
    #  Step 1. Get pressure-weighted u-component wind
    pressure_weighted_ua = jet_statistics_components.calc_mass_weighted_average(
        data, ws_col="ua"
    )
    #  Step 2. Filter pressure-weighted u-component wind with low-pass Lanczos filter
    filtered_pressure_weighted_ua = jet_statistics_components.apply_lanczos_filter(
        pressure_weighted_ua,
        filter_freq=filter_freq,
        window_size=window_size,
    )
    #  Step 3. Turn dataarray into dataset for next part
    filtered_pressure_weighted_ua = filtered_pressure_weighted_ua.to_dataset(name="ua")

    #  Step 4.  Get max latitude and wind speed at max
    zonal_mean = windspeed_utils.get_zonal_mean(filtered_pressure_weighted_ua)
    all_max_lats_and_ws = np.array(
        list(
            map(
                jet_statistics_components.get_3_latitudes_and_speed_around_max_ws,
                zonal_mean["ua"],
            )
        )
    )

    #  Step 5. Scale max latitude to 0.01 degree using quadratic function
    scaled_max_lats = []
    scaled_max_ws = []
    for max_lat_and_ws in all_max_lats_and_ws:
        if not np.isnan(np.min(max_lat_and_ws)):
            (
                scaled_max_lat,
                scaled_ws,
            ) = jet_statistics_components.get_latitude_and_speed_where_max_ws_at_reduced_resolution(
                max_lat_and_ws, lat_resolution=0.01
            )
            scaled_max_lats.append(scaled_max_lat)
            scaled_max_ws.append(scaled_ws)
        else:
            if np.isnan(max_lat_and_ws).all():
                scaled_max_lats.append(np.nan)
                scaled_max_ws.append(np.nan)
            else:
                if np.nanmax(max_lat_and_ws[0]) == data["lat"].max():
                    scaled_max_lats.append(np.nanmax(max_lat_and_ws[0]))
                    scaled_max_ws.append(np.nanmax(max_lat_and_ws[1]))
                elif np.nanmin(max_lat_and_ws[0]) == data["lat"].min():
                    scaled_max_lats.append(np.nanmin(max_lat_and_ws[0]))
                    scaled_max_ws.append(np.nanmin(max_lat_and_ws[1]))
                else:
                    scaled_max_lats.append(np.nan)
                    scaled_max_ws.append(np.nan)

    #  Step 6. Get jet-widths using scaled windspeed and usual jet-lat
    max_lats = all_max_lats_and_ws[::, 0, 1]
    jet_widths = list(
        map(
            lambda zm, la, wa: jet_statistics_components.calc_jet_width_for_one_day(
                zm, la, wa
            ),
            zonal_mean["ua"],
            max_lats,
            scaled_max_ws,
        )
    )

    #  Step 7. Assign the new output variables to the original data
    output = data.assign(
        {
            "jet_lat": (("time"), scaled_max_lats),
            "jet_speed": (("time"), scaled_max_ws),
            "jet_width": (("time"), jet_widths),
        }
    )
    return output


@sort_xarray_data_coords(coords=["lat", "lon"])
def grise_polvani_2014(data):
    r"""
    This method calculates the latitude of the midlatitude eddy-driven jet ('jet_lat') by finding the peak value of the input u-wind field.
    A polynomial fit is then applied to get an appropriate value of 'jet_lat' at a resolution 0.01 degrees.
    As opposed to the original method, this implementation also returns the speed at the 'jet_lat': the 'jet_speed'

    This method was originally introduce in Grise & Polvani (2014) https://doi.org/10.1002/2013GL058466
    and is described in Section 2 of that study.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua', and the coordinates: 'lon', 'lat', 'plev' and 'time'.

    Returns
    ----------
    output : xarray.Dataset
        Data containing the two outputs: 'jet_lat' and 'jet_speed'

    Notes
    -----
    This method was originally developed for the jet streams in the Southern Hemisphere.

    The original paper also includes two other metrics for zonal mean atmospheric circulation.

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u component wind:
        ua_data = xr.open_dataset('path_to_u_data')

        # Subset dataset to range used in original methodology (850 hPa &  -65--30N)):
        ua_sub = ua.sel(plev=850, lat=slice(-65, -30))

        # Run statistic:
        gp16 = jsmetrics.jet_statistics.grise_polvani_2014(ua_sub)
    """
    if isinstance(data, xarray.DataArray):
        data = data.to_dataset()

    # Step 1. Expand time dimensions so we can map a function to the dataset properly
    if "time" not in data.coords:
        raise KeyError("Please provide a time coordinate for data to run this metric")
    if data["time"].size == 1 and "time" not in data.dims:
        data = data.expand_dims("time")

    # Step 2. Calculate zonal-mean
    zonal_mean = windspeed_utils.get_zonal_mean(data)

    # Step 3. Get the 3 latitudes and speeds around max zonal wind-speed (e.g. lat-1, lat, lat+1)
    all_max_lats_and_ws = np.array(
        list(
            map(
                jet_statistics_components.get_3_latitudes_and_speed_around_max_ws,
                zonal_mean["ua"],
            )
        )
    )

    #  Step 4. Apply quadratic function to get max latitude at 0.01 degree resolution
    scaled_max_lats = []
    scaled_max_ws = []
    for max_lat_and_ws in all_max_lats_and_ws:
        try:
            (
                scaled_max_lat,
                scaled_ws,
            ) = jet_statistics_components.get_latitude_and_speed_where_max_ws_at_reduced_resolution(
                max_lat_and_ws, lat_resolution=0.01
            )
        except Exception as e:
            print(e)
            scaled_max_lat = np.nan
            scaled_ws = np.nan
        scaled_max_lats.append(scaled_max_lat)
        scaled_max_ws.append(scaled_ws)

    #  Step 5. Assign scaled max lats back to data
    output = data.assign(
        {
            "jet_lat": (("time"), scaled_max_lats),
            "jet_speed": (("time"), scaled_max_ws),
        }
    )
    return output


@sort_xarray_data_coords(coords=["lat", "lon"])
def barnes_polvani_2015(data):
    r"""
    This method calculates the jet positon and wind speed at that position by fitting a parabola around the
    maximum of zonally average u-component wind speed, using the magnitude at the maximum ('jet_speed') and latitude at that maximum ('jet_lat').

    This method was originally introduce in Barnes & Polvani (2015) https://doi.org/10.1175/JCLI-D-14-00589.1
    and is described in Section 2b of that study.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua', and the coordinates: 'lon', 'lat', 'plev' and 'time'.

    Returns
    ----------
    output : xarray.Dataset
        Data containing the two outputs: 'jet_lat' and 'jet_speed'

    Notes
    -----
    This methodology make an assumption that the a parabola can be fit to windspeed profile, so it performs quite different from
    other jet latitude methods available in the package in cases where the windspeed profile is more complex (multiple jets),
    and on data with finer temporal resolution in general.

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u component wind:
        ua_data = xr.open_dataset('path_to_u_data')

        # Subset dataset to range used in original methodology (700-925 hPa &  30-70N,  130-10W)):
        ua_sub = ua.sel(plev=slice(700, 925), lat=slice(30, 70), lon=slice(230, 350))

        # Run statistic:
        bp15 = jsmetrics.jet_statistics.barnes_polvani_2015(ua_sub)
    """
    # Step 1. Get zonal mean
    zonal_mean = windspeed_utils.get_zonal_mean(data)

    # Step 2. Get jet lat and jet speed values and assign to output data
    if zonal_mean["time"].size == 1:
        if "time" in zonal_mean.dims:
            zonal_mean = zonal_mean.squeeze("time")
        output = jet_statistics_components.get_jet_lat_and_speed_using_parabola_by_day(
            zonal_mean
        )
    else:
        output = zonal_mean.groupby("time").map(
            jet_statistics_components.get_jet_lat_and_speed_using_parabola_by_day
        )
    return output


@sort_xarray_data_coords(coords=["lat", "lon"])
def barnes_simpson_2017(data):
    r"""
    This method calculates two outputs: 'jet_lat' and 'jet_speed' which are defined as the latitude and speed of the 10-day-averaged
    maximum zonally-averaged wind speed.

    This method was originally introduce in Barnes & Simpson 2017 https://doi.org/10.1175/JCLI-D-17-0299.1
    and is described in Section 2b of that study.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua', and the coordinates: 'lon', 'lat', 'plev' and 'time'.

    Returns
    ----------
    output : xarray.Dataset
        Data containing the x outputs: 'jet_lat' and 'jet_speed'

    Notes
    -----
    The original methodology was intended to work on one pressure level (700 hPa) and on daily data, for the
    implementation included in this package, we have included methods to automatically average any inputted pressure levels
    and to return the data without 10-day averaging if data above the 10-day resolution is inputted.
    Instead, warnings are returned to user to help them use the method the way it was originally intended.

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u component wind:
        ua_data = xr.open_dataset('path_to_u_data')

        # Subset dataset to range used in original methodology (700 hPa & North Atlantic & North Pacific)):
        ua_na = ua.sel(plev=700, lat=slice(0, 90), lon=slice(280, 350)) # North Atlantic
        ua_np = ua.sel(plev=700, lat=slice(0, 90), lon=slice(120, 230)) # North Pacific

        # Run statistic:
        bp17_na = jsmetrics.jet_statistics.barnes_simpson_2017(ua_na)
        bp17_np = jsmetrics.jet_statistics.barnes_simpson_2017(ua_np)

    """
    # Step 1. Run checks on the 'plev' coordinate. There should only be 1 for this method, but this implementation allows for multiple.
    if "plev" in data.dims:
        if data["plev"].count() == 1:
            data = data.isel(plev=0)
        else:
            print(
                "this metric was meant to only work on one plev, please subset plev to one value. For now taking the mean..."
            )
            data = data.mean("plev")
    data = data.mean("lon")

    # Step 1. Run check on time coordinate. The data should be able to be resampled into 10 days for this method.
    # Check 1. Is time coordinate is in data and is monotonically increasing.
    if "time" not in data.coords:
        raise KeyError("Please provide a time coordinate for data to run this metric")
    if data["time"].size == 1 and "time" not in data.dims:
        data = data.expand_dims("time")
    if not data.indexes["time"].is_monotonic_increasing:
        raise IndexError("Data needs to have a montonic increasing index")
    # Check 2. Can that data be resampled into 10 days
    if not data["time"].size == 1:
        time_step_in_data = int((data["time"][1] - data["time"][0]).dt.days)
        if time_step_in_data <= 10:
            data = data.resample(time="10D").mean()
            time_step_in_data = 10
        else:
            print(
                f"Warning this method was developed for 10 day average and data has larger time-step than 10 days. Time step is {time_step_in_data} days"
            )
    data = data.dropna("time")  # Drop all NaN slices

    # Step 3. Calculate jet lat and jet speed.

    data = jet_statistics_components.calc_latitude_and_speed_where_max_ws(data)
    return data


@sort_xarray_data_coords(coords=["lat", "lon"])
def bracegirdle_et_al_2018(data):
    r"""
    This method calculates the seasonal and annual jet-stream position ('JPOS') and strength ('JSTR')
    by applying a 0.075 degrees cubic spline interpolation to a zonally-averaged wind climatology
    and selecting the maximum.

    This method was originally introduce in Bracegirdle et al (2018) https://doi.org/10.1175/JCLI-D-17-0320.1
    and is described in Section 2 of that study.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua', and the coordinates: 'lon', 'lat', 'plev' and 'time'.

    Returns
    ----------
    output : xarray.Dataset
        Data containing the four outputs: 'annual_JPOS', 'seasonal_JPOS', 'annual_JSTR' and 'seasonal_JSTR'

    Notes
    -----
    This method was originally developed for the jet streams in the Southern Hemisphere.

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u component wind:
        ua_data = xr.open_dataset('path_to_u_data')

        # Subset dataset to range used in original methodology (850 hPa &  -75--10N)):
        ua_sub = ua.sel(plev=850, lat=slice(-75, -10))

        # Run statistic:
        b18 = jsmetrics.jet_statistics.bracegirdle_et_al_2018(ua_sub)
    """
    # Checks on plev
    if isinstance(data, xarray.DataArray):
        data = data.to_dataset()
    if "plev" in data.dims:
        if data["plev"].count() == 1:
            data = data.isel(plev=0)
        else:
            print(
                "this metric was meant to only work on one plev, please subset plev to one value. For now taking the mean..."
            )
            data = data.mean("plev")
            # raise ValueError("Please subset to one plev value for this metric")

    # Step 1: Expand time dimensions so we can map a function to the dataset properly
    if "time" not in data.coords:
        raise KeyError("Please provide a time coordinate for data to run this metric")
    if data["time"].size == 1 and "time" not in data.dims:
        data = data.expand_dims("time")

    #  Step 2. Calculate seasonal & annual climatologies
    seasonal_climatology = jet_statistics_components.get_climatology(data, "season")
    annual_climatology = jet_statistics_components.get_climatology(data, "year")

    #  Step 3. Get zonal mean from climatologies
    seasonal_zonal_mean = seasonal_climatology.mean("lon")
    annual_zonal_mean = annual_climatology.mean("lon")

    #  Step 4. Cubic spline interpolation to each climatology at latitude resolution of 0.075 degrees
    (
        seasonal_max_lats,
        seasonal_max_ws,
    ) = jet_statistics_components.run_cubic_spline_interpolation_for_each_unit_of_climatology_to_get_max_lat_and_ws(
        seasonal_zonal_mean, lat_resolution=0.075, time_col="season"
    )
    (
        annual_max_lats,
        annual_max_ws,
    ) = jet_statistics_components.run_cubic_spline_interpolation_for_each_unit_of_climatology_to_get_max_lat_and_ws(
        annual_zonal_mean, lat_resolution=0.075, time_col="year"
    )

    # Step 5. Assign jet-stream position (JPOS) and jet-stream strength (JSTR) back to data
    output = data.assign(
        {
            "seasonal_JPOS": (("season"), seasonal_max_lats),
            "annual_JPOS": (("year"), annual_max_lats),
            "seasonal_JSTR": (("season"), seasonal_max_ws),
            "annual_JSTR": (("year"), annual_max_ws),
        }
    )
    return output


@sort_xarray_data_coords(coords=["lat", "lon"])
def ceppi_et_al_2018(data, lon_resolution=None):
    r"""
    This method calculates the jet latitude ('jet-lat') as defined by selecting the centroid of a zonally-averaged wind profile.

    The centroid is calculated by:

    .. math::
        \phi_{jet}  = \frac{\int_{30°}^{60°} \phi\bar{u}^2, d\phi}{\int_{30°}^{60°} \bar{u}^2, d\phi}

    This method has been slightly adapted to include a 'jet_speed' extraction (provided for this method
    in Screen et al. (2022) https://doi.org/10.1029/2022GL100523).

    **Note:** The implementation here does not explicit limit the centroid calculation to latitude between 20°-70°,
    instead this range is determined by the input data.

    This method was originally introduce in Ceppi et al (2018) https://doi.org/10.1175/JCLI-D-17-0323.1
    and is described in Section 2b of that study.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua', and the coordinates: 'lon', 'lat', 'plev' and 'time'.
    lon_resolution : numeric
        Resolution to use for longitude coord if size 1

    Returns
    ----------
    output : xarray.Dataset
        Data containing the three outputs: 'jet_lat', 'jet_speed', 'total_area_m2'

    Notes
    -----
    This method was improved by Zappa et al. (2018) https://doi.org/10.1029/2019GL083653,
    which includes exclusion of :math:`<0 m s^{-1}` u-wind. This method is also available in jsmetrics.

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u component wind:
        ua_data = xr.open_dataset('path_to_u_data')

        # Subset dataset to range used in original methodology (850 hPa &  30-60N/S)):
        ua_na = ua.sel(plev=850, lat=slice(30, 60), lon=slice(300, 60)) # North Atlantic-European Sector
        ua_na = ua.sel(plev=850, lat=slice(30, 60), lon=slice(140, 240)) # North Pacific
        ua_sh = ua.sel(plev=850, lat=slice(-60, -30)) # Southern Hemisphere

        # Run statistic:
        c18_na = jsmetrics.jet_statistics.ceppi_et_al_2018(ua_na)
        c18_np = jsmetrics.jet_statistics.ceppi_et_al_2018(ua_np)
        c18_sh = jsmetrics.jet_statistics.ceppi_et_al_2018(ua_sh)
    """
    #  Step 1. Get area in m2 by latitude/longitude grid cells
    if not data["lon"].size == 1 and not data["lat"].size == 1:
        total_area_m2 = spatial_utils.grid_cell_areas(data["lon"], data["lat"])
    elif lon_resolution and not data["lat"].size == 1 and data["lon"].size == 1:
        lons_to_use = [float(data["lon"]), float(data["lon"]) + lon_resolution]
        total_area_m2 = spatial_utils.grid_cell_areas(lons_to_use, data["lat"])
        total_area_m2 = total_area_m2.mean(axis=1)
        total_area_m2 = total_area_m2.reshape(-1, 1)
        if data["lon"].shape == ():
            data = data.expand_dims("lon")
    else:
        raise ValueError(
            "For this method, your data needs to have at least 2 'lat' values and 'lon' values needs to be more than one unless you set the 'lon_resolution' parameter"
        )

    data["total_area_m2"] = (("lat", "lon"), total_area_m2)

    #  Step 2. calculate zonal mean
    zonal_mean = windspeed_utils.get_zonal_mean(data)

    # Step 3: Assign laitude of jet-stream centroids to main data
    data["jet_lat"] = jet_statistics_components.calc_centroid_jet_lat_from_zonal_mean(
        zonal_mean, area_by_lat=zonal_mean["total_area_m2"]
    )

    # Step 4. Expand time dimension
    if "time" not in data.coords:
        raise KeyError("Please provide a time coordinate for data to run this metric")
    if data["time"].size == 1 and "time" not in data.dims:
        data = data.expand_dims("time")
        zonal_mean = zonal_mean.expand_dims("time")

    # Step 5 (adapted from original methodology): Get nearest latitude actually in data to the one determined by metric
    nearest_latitudes_to_jet_lat_estimates = np.array(
        list(
            map(
                lambda row: data_utils.find_nearest_value_to_array(
                    data["lat"], float(row)
                ),
                data["jet_lat"],
            )
        )
    )

    # Step 6 (adapted from original methodology): Get speed of associated nearest latitude
    data["jet_speed"] = (
        ("time",),
        np.array(
            list(
                map(
                    lambda data_row, lat_val: jet_statistics_components.get_latitude_value_in_data_row(
                        data_row, lat_val
                    ),
                    zonal_mean["ua"],
                    nearest_latitudes_to_jet_lat_estimates,
                )
            )
        ),
    )
    return data


def zappa_et_al_2018(data, lon_resolution=None):
    r"""
    This method calculates the jet latitude ('jet-lat') as defined by selecting the centroid of a zonally-averaged wind profile.

    The centroid is calculated by:

    .. math::
        \phi_{jet}  = \frac{\int_{20°}^{70°} \phi\bar{u}^2_0, d\phi}{\int_{20°}^{70°} \bar{u}^2_0, d\phi}

    .. math::
        u_0(\phi) = \max(0, u(\phi))

    **Note:** The implementation here does not explicit limit the centroid calculation to latitude between 20°-70°,
    instead this range is determined by the input data.

    This method has been slightly adapted to include a 'jet_speed' extraction (provided for this method
    in Screen et al. (2022) https://doi.org/10.1029/2022GL100523).

    This method was originally introduced in Zappa et al. 2018 https://doi.org/10.1029/2019GL083653
    and is described in Section 2.3 of that study.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua', and the coordinates: 'lon', 'lat', 'plev' and 'time'.
    lon_resolution : numeric
        Resolution to use for longitude coord if size 1

    Returns
    ----------
    output : xarray.Dataset
        Data containing the three outputs: 'jet_lat', 'jet_speed', 'total_area_m2'

    Notes
    -----
    This method was adapted from and very similar to Ceppi et al (2018) https://doi.org/10.1175/JCLI-D-17-0323.1.

    This method is also used in Ayres & Screen, 2019 and Screen et al. 2022. Similar methods are used in:
    Chen et al. 2008; Ceppi et al. 2014, Ceppi et al. 2018.

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u component wind:
        ua_data = xr.open_dataset('path_to_u_data')

        # Subset dataset to range used in original methodology (850 hPa &  20-70N, 140∘E-240∘E or 300-360 E))):
        ua_na = ua.sel(plev=850, lat=slice(20, 70), lon=slice(140, 240))
        ua_np = ua.sel(plev=850, lat=slice(20, 70), lon=slice(300, 360))


        # Run statistic:
        z18_na = jsmetrics.jet_statistics.zappa_et_al_2018(ua_na)
        z18_np = jsmetrics.jet_statistics.zappa_et_al_2018(ua_np)
    """
    #  Step 1. Get area in m2 by latitude/longitude grid cells
    if not data["lon"].size == 1 and not data["lat"].size == 1:
        total_area_m2 = spatial_utils.grid_cell_areas(data["lon"], data["lat"])
    elif lon_resolution and not data["lat"].size == 1 and data["lon"].size == 1:
        lons_to_use = [float(data["lon"]), float(data["lon"]) + lon_resolution]
        total_area_m2 = spatial_utils.grid_cell_areas(lons_to_use, data["lat"])
        total_area_m2 = total_area_m2.mean(axis=1)
        total_area_m2 = total_area_m2.reshape(-1, 1)
        if data["lon"].shape == ():
            data = data.expand_dims("lon")
    else:
        raise ValueError(
            "For this method, your data needs to have at least 2 'lat' values and 'lon' values needs to be more than one unless you set the 'lon_resolution' parameter"
        )

    data["total_area_m2"] = (("lat", "lon"), total_area_m2)

    #  Step 2. calculate zonal mean and floor ua values to 0
    zonal_mean = windspeed_utils.get_zonal_mean(data)
    zonal_mean["ua"] = zonal_mean["ua"].where(lambda x: x > 0, 0)

    # Step 3: Assign laitude of jet-stream centroids to main data
    data["jet_lat"] = jet_statistics_components.calc_centroid_jet_lat_from_zonal_mean(
        zonal_mean, area_by_lat=zonal_mean["total_area_m2"]
    )

    # Step 4. Expand time dimension
    if "time" not in data.coords:
        raise KeyError("Please provide a time coordinate for data to run this metric")
    if data["time"].size == 1 and "time" not in data.dims:
        data = data.expand_dims("time")
        zonal_mean = zonal_mean.expand_dims("time")

    # Step 5 (adapted from original methodology): Get nearest latitude actually in data to the one determined by metric
    nearest_latitudes_to_jet_lat_estimates = np.array(
        list(
            map(
                lambda row: data_utils.find_nearest_value_to_array(
                    data["lat"], float(row)
                ),
                data["jet_lat"],
            )
        )
    )

    # Step 6 (adapted from original methodology): Get speed of associated nearest latitude
    data["jet_speed"] = (
        ("time",),
        np.array(
            list(
                map(
                    lambda data_row, lat_val: jet_statistics_components.get_latitude_value_in_data_row(
                        data_row, lat_val
                    ),
                    zonal_mean["ua"],
                    nearest_latitudes_to_jet_lat_estimates,
                )
            )
        ),
    )
    return data


@sort_xarray_data_coords(coords=["lat", "lon"])
def kerr_et_al_2020(data, width_of_pulse=10):
    r"""
    This method defines the latitude and speed of the jet-stream where the maximum zonal winds occur for
    each longitude and for each time unit (i.e. day). These values are then smoothed across the longitudes
    with a rectangular pulse (by default this has a width of 10 degrees).

    This method was originally introduced in Kerr et al. (2020) https://onlinelibrary.wiley.com/doi/10.1029/2020JD032735
    and is described in Section 2.4.2 of that study.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua', and the coordinates: 'lon', 'lat', 'plev' and 'time'.

    Returns
    ----------
    output : xarray.Dataset
        Data containing the two outputs: 'jet_lat' and 'smoothed_jet_lat'

    Notes
    -----
    This method was based on the method from Barnes and Fiore (2013) https://doi.org/10.1002/grl.50411

    The implementation here returns both smoothed and unsmoothed jet latitude outputs.

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u component wind:
        ua_data = xr.open_dataset('path_to_u_data')

        # Subset dataset to range used in original methodology (700 hPa &  20-70N,  300-360W)):
        ua_sub = ua.sel(plev=700, lat=slice(20, 70), lon=slice(300, 360))

        # Run statistic:
        k20 = jsmetrics.jet_statistics.kerr_et_al_2020(ua_sub)
    """
    # Checks on plev coordinate
    if "plev" in data.dims:
        if data["plev"].count() == 1:
            data = data.isel(plev=0)
        else:
            print(
                "this metric was meant to only work on one plev, please subset plev to one value. For now taking the mean..."
            )
            data = data.mean("plev")

    # Step 1. Calculateed smoothed jet lats by lon
    if "time" not in data.coords:
        raise KeyError("Please provide a time coordinate for data to run this metric")
    elif data["time"].size == 1:
        if "time" in data.dims:
            data = data.squeeze("time")
        output = (
            jet_statistics_components.get_moving_averaged_smoothed_jet_lats_for_one_day(
                data, width_of_pulse
            )
        )
    else:
        output = data.groupby("time").map(
            jet_statistics_components.get_moving_averaged_smoothed_jet_lats_for_one_day,
            (width_of_pulse,),
        )
    return output
